from django.conf import settings
from django.contrib import messages
from django.utils.html import strip_tags
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext
from django.views import generic

from oscar.core.loading import get_class, get_model
from oscar.core.utils import redirect_to_referrer
from oscar.views.generic import BulkEditMixin

PageTitleMixin = get_class("customer.mixins", "PageTitleMixin")
Notification = get_model("communication", "Notification")


class NotificationListView(PageTitleMixin, generic.ListView):
    model = Notification
    template_name = "oscar/communication/notifications/list.html"
    context_object_name = "notifications"
    paginate_by = settings.OSCAR_NOTIFICATIONS_PER_PAGE
    page_title = _("Notifications")
    active_tab = "notifications"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["list_type"] = self.list_type
        return ctx


class InboxView(NotificationListView):
    list_type = "inbox"

    def get_queryset(self):
        return self.model._default_manager.filter(
            recipient=self.request.user, location=self.model.INBOX
        )


class ArchiveView(NotificationListView):
    list_type = "archive"

    def get_queryset(self):
        return self.model._default_manager.filter(
            recipient=self.request.user, location=self.model.ARCHIVE
        )


class DetailView(PageTitleMixin, generic.DetailView):
    model = Notification
    template_name = "oscar/communication/notifications/detail.html"
    context_object_name = "notification"
    active_tab = "notifications"

    def get_object(self, queryset=None):
        obj = super().get_object()
        if not obj.date_read:
            obj.date_read = now()
            obj.save()
        return obj

    def get_page_title(self):
        """Append subject to page title"""
        title = strip_tags(self.object.subject)
        return "%s: %s" % (_("Notification"), title)

    def get_queryset(self):
        return self.model._default_manager.filter(recipient=self.request.user)


class UpdateView(BulkEditMixin, generic.View):
    model = Notification
    http_method_names = ["post"]
    actions = ("archive", "delete")
    checkbox_object_name = "notification"

    def get_object_dict(self, ids):
        return self.model.objects.filter(recipient=self.request.user).in_bulk(ids)

    def get_success_response(self):
        return redirect_to_referrer(self.request, "communication:notifications-inbox")

    def archive(self, request, notifications):
        for notification in notifications:
            notification.archive()
        msg = ngettext(
            "%(count)d notification archived",
            "%(count)d notifications archived",
            len(notifications),
        ) % {"count": len(notifications)}
        messages.success(request, msg)
        return self.get_success_response()

    def delete(self, request, notifications):
        for notification in notifications:
            notification.delete()
        msg = ngettext(
            "%(count)d notification deleted",
            "%(count)d notifications deleted",
            len(notifications),
        ) % {"count": len(notifications)}
        messages.success(request, msg)
        return self.get_success_response()



from io import TextIOWrapper

from django.conf import settings
from django.contrib import messages
from django.core import exceptions
from django.db.models import Count
from django.shortcuts import HttpResponse, get_object_or_404, redirect
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext
from django.views.generic import CreateView, DeleteView, ListView, UpdateView, View

from oscar.core.loading import get_classes, get_model
from oscar.views.generic import BulkEditMixin

Range = get_model("offer", "Range")
RangeProduct = get_model("offer", "RangeProduct")
RangeProductFileUpload = get_model("offer", "RangeProductFileUpload")
Product = get_model("catalogue", "Product")
RangeForm, RangeProductForm = get_classes(
    "dashboard.ranges.forms", ["RangeForm", "RangeProductForm"]
)


class RangeListView(ListView):
    model = Range
    context_object_name = "ranges"
    template_name = "oscar/dashboard/ranges/range_list.html"
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    def get_queryset(self):
        return self.model._default_manager.prefetch_related("included_categories")


class RangeCreateView(CreateView):
    model = Range
    template_name = "oscar/dashboard/ranges/range_form.html"
    form_class = RangeForm

    def get_success_url(self):
        if "action" in self.request.POST:
            return reverse("dashboard:range-products", kwargs={"pk": self.object.id})
        else:
            msg = render_to_string(
                "oscar/dashboard/ranges/messages/range_saved.html",
                {"range": self.object},
            )
            messages.success(self.request, msg, extra_tags="safe noicon")
            return reverse("dashboard:range-list")

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Create range")
        return ctx


class RangeUpdateView(UpdateView):
    model = Range
    template_name = "oscar/dashboard/ranges/range_form.html"
    form_class = RangeForm

    def get_object(self, queryset=None):
        obj = super().get_object(queryset)
        if not obj.is_editable:
            raise exceptions.PermissionDenied("Not allowed")
        return obj

    def get_success_url(self):
        if "action" in self.request.POST:
            return reverse("dashboard:range-products", kwargs={"pk": self.object.id})
        else:
            msg = render_to_string(
                "oscar/dashboard/ranges/messages/range_saved.html",
                {"range": self.object},
            )
            messages.success(self.request, msg, extra_tags="safe noicon")
            return reverse("dashboard:range-list")

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["range"] = self.object
        ctx["title"] = self.object.name
        return ctx


class RangeDeleteView(DeleteView):
    model = Range
    template_name = "oscar/dashboard/ranges/range_delete.html"
    context_object_name = "range"

    def get_success_url(self):
        messages.warning(self.request, _("Range deleted"))
        return reverse("dashboard:range-list")


class RangeProductListView(BulkEditMixin, ListView):
    model = Product
    template_name = "oscar/dashboard/ranges/range_product_list.html"
    context_object_name = "products"
    actions = (
        "add_products",
        "add_excluded_products",
        "remove_selected_products",
        "remove_excluded_products",
    )
    form_class = RangeProductForm
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE

    # pylint: disable=attribute-defined-outside-init
    def get(self, request, *args, **kwargs):
        self.upload_type = request.GET.get("upload_type", "")
        return super().get(request, *args, **kwargs)

    # pylint: disable=attribute-defined-outside-init
    def post(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        self.upload_type = request.POST.get("upload_type", "")
        if request.POST.get("action", None) == "add_products":
            return self.add_products(request)
        if request.POST.get("action", None) == "add_excluded_products":
            return self.add_excluded_products(request)
        if request.POST.get("action", None) == "remove_excluded_products":
            return self.remove_excluded_products(request)
        return super().post(request, *args, **kwargs)

    def get_product_range(self):
        if not hasattr(self, "_product_range"):
            self._product_range = get_object_or_404(Range, id=self.kwargs["pk"])
        return self._product_range

    def get_queryset(self):
        products = self.get_product_range().all_products()
        return products.order_by("rangeproduct__display_order")

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        product_range = self.get_product_range()
        ctx["range"] = product_range
        if "form" not in ctx:
            ctx["form"] = self.form_class(
                product_range,
                initial={"upload_type": RangeProductFileUpload.INCLUDED_PRODUCTS_TYPE},
            )
        if "form_excluded" not in ctx:
            ctx["form_excluded"] = self.form_class(
                product_range,
                initial={"upload_type": RangeProductFileUpload.EXCLUDED_PRODUCTS_TYPE},
            )
        ctx["file_uploads_included"] = product_range.file_uploads.filter(
            upload_type=RangeProductFileUpload.INCLUDED_PRODUCTS_TYPE
        )
        ctx["file_uploads_excluded"] = product_range.file_uploads.filter(
            upload_type=RangeProductFileUpload.EXCLUDED_PRODUCTS_TYPE
        )
        ctx["upload_type"] = self.upload_type
        return ctx

    def remove_selected_products(self, request, products):
        product_range = self.get_product_range()
        for product in products:
            product_range.remove_product(product)
        num_products = len(products)
        messages.success(
            request,
            ngettext(
                "Removed %d product from range",
                "Removed %d products from range",
                num_products,
            )
            % num_products,
        )
        return redirect(
            reverse("dashboard:range-products", kwargs={"pk": product_range.pk})
        )

    def add_products(self, request):
        product_range = self.get_product_range()
        form = self.form_class(product_range, request.POST, request.FILES)
        if not form.is_valid():
            ctx = self.get_context_data(form=form, object_list=self.object_list)
            return self.render_to_response(ctx)

        self.handle_query_products(request, product_range, form)
        self.handle_file_products(request, product_range, form)
        return redirect(
            reverse("dashboard:range-products", kwargs={"pk": product_range.pk})
        )

    def add_excluded_products(self, request):
        product_range = self.get_product_range()
        form = self.form_class(
            product_range,
            request.POST,
            request.FILES,
            initial={"upload_type": RangeProductFileUpload.EXCLUDED_PRODUCTS_TYPE},
        )
        if not form.is_valid():
            ctx = self.get_context_data(
                form_excluded=form, object_list=self.object_list
            )
            return self.render_to_response(ctx)

        self.handle_query_products(request, product_range, form)
        self.handle_file_products(request, product_range, form)
        return redirect(
            reverse("dashboard:range-products", kwargs={"pk": product_range.pk})
            + "?upload_type=excluded"
        )

    def remove_excluded_products(self, request):
        product_ids = request.POST.getlist("selected_product", None)
        products = self.model.objects.filter(id__in=product_ids)
        product_range = self.get_product_range()
        for product in products:
            product_range.excluded_products.remove(product)
        num_products = len(products)
        messages.success(
            request,
            ngettext(
                "Removed %d product from excluded list",
                "Removed %d products from excluded list",
                num_products,
            )
            % num_products,
        )
        return redirect(
            reverse("dashboard:range-products", kwargs={"pk": product_range.pk})
            + "?upload_type=excluded"
        )

    def handle_query_products(self, request, product_range, form):
        products = form.get_products()
        if not products:
            return
        for product in products:
            if (
                form.cleaned_data["upload_type"]
                == RangeProductFileUpload.EXCLUDED_PRODUCTS_TYPE
            ):
                product_range.excluded_products.add(product)
                action = _("excluded from this range")
            else:
                product_range.add_product(product)
                action = _("added to this range")
        num_products = len(products)
        messages.success(
            request,
            ngettext(
                "%(num_products)d product has been %(action)s",
                "%(num_products)d products have been %(action)s",
                num_products,
            )
            % {"num_products": num_products, "action": action},
        )
        dupe_skus = form.get_duplicate_skus()
        if dupe_skus:
            messages.warning(
                request,
                _(
                    "The products with SKUs or UPCs matching %(skus)s have "
                    "already been %(action)s"
                )
                % {"skus": ", ".join(dupe_skus), "action": action},
            )

        missing_skus = form.get_missing_skus()
        if missing_skus:
            messages.warning(
                request,
                _("No product(s) were found with SKU or UPC matching %s")
                % ", ".join(missing_skus),
            )
        self.check_imported_products_sku_duplicates(request, products)

    def handle_file_products(self, request, product_range, form):
        if "file_upload" not in request.FILES:
            return
        f = request.FILES["file_upload"]
        upload = self.create_upload_object(
            request, product_range, f, form.cleaned_data["upload_type"]
        )
        products = upload.process(TextIOWrapper(f, encoding=request.encoding))
        if not upload.was_processing_successful():
            messages.error(request, upload.error_message)
        else:
            if (
                form.cleaned_data["upload_type"]
                == RangeProductFileUpload.EXCLUDED_PRODUCTS_TYPE
            ):
                action = "excluded from this range"
            else:
                action = "added to this range"
            msg = render_to_string(
                "oscar/dashboard/ranges/messages/range_products_saved.html",
                {"range": product_range, "upload": upload, "action": action},
            )
            messages.success(request, msg, extra_tags="safe noicon block")
        self.check_imported_products_sku_duplicates(request, products)

    def create_upload_object(self, request, product_range, f, upload_type):
        upload = RangeProductFileUpload.objects.create(
            range=product_range,
            uploaded_by=request.user,
            filepath=f.name,
            size=f.size,
            upload_type=upload_type,
        )
        return upload

    def check_imported_products_sku_duplicates(self, request, queryset):
        dupe_sku_products = (
            queryset.values("stockrecords__partner_sku")
            .annotate(total=Count("stockrecords__partner_sku"))
            .filter(total__gt=1)
            .order_by("stockrecords__partner_sku")
        )
        if dupe_sku_products:
            dupe_skus = [p["stockrecords__partner_sku"] for p in dupe_sku_products]
            messages.warning(
                request,
                _("There are more than one product with SKU %s") % ", ".join(dupe_skus),
            )


class RangeReorderView(View):
    # pylint: disable=unused-argument
    def post(self, request, pk):
        order = dict(request.POST).get("product", [])
        self._save_page_order(order)
        return HttpResponse(status=200)

    def _save_page_order(self, order):
        """
        Save the order of the products within range.
        """
        range_products = RangeProduct.objects.filter(
            range_id=self.kwargs["pk"], product_id__in=order
        )
        for range_product in range_products:
            range_product.display_order = order.index(str(range_product.product_id))
        RangeProduct.objects.bulk_update(range_products, ["display_order"])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-03-27 07:57
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0015_product_is_public'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_boolean',
            field=models.NullBooleanField(db_index=True, verbose_name='Boolean'),
        ),
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_date',
            field=models.DateField(blank=True, db_index=True, null=True, verbose_name='Date'),
        ),
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_datetime',
            field=models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='DateTime'),
        ),
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_float',
            field=models.FloatField(blank=True, db_index=True, null=True, verbose_name='Float'),
        ),
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_integer',
            field=models.IntegerField(blank=True, db_index=True, null=True, verbose_name='Integer'),
        ),
    ]



from django.db.models.signals import post_delete
from django.dispatch import receiver

from oscar.core.loading import get_model

ConditionalOffer = get_model("offer", "ConditionalOffer")
Condition = get_model("offer", "Condition")
Benefit = get_model("offer", "Benefit")


@receiver(post_delete, sender=ConditionalOffer)
def delete_unused_related_conditions_and_benefits(instance, **kwargs):
    offer = instance  # the object is no longer in the database

    try:
        condition = Condition.objects.get(id=offer.condition_id)
    except Condition.DoesNotExist:
        pass
    else:
        # Only delete if not using a proxy, and not used by other offers
        if condition.proxy_class == "" and not condition.offers.exists():
            condition.delete()

    try:
        benefit = Benefit.objects.get(id=offer.benefit_id)
    except Benefit.DoesNotExist:
        pass
    else:
        # Only delete if not using a proxy, and not used by other offers
        if benefit.proxy_class == "" and not benefit.offers.exists():
            benefit.delete()



# Generated by Django 4.2.5 on 2023-09-15 11:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("offer", "0011_rangeproductfileupload_included"),
    ]

    operations = [
        migrations.CreateModel(
            name="FixedUnitDiscountBenefit",
            fields=[],
            options={
                "verbose_name": "Fixed unit discount benefit",
                "verbose_name_plural": "Fixed unit discount benefits",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("offer.absolutediscountbenefit",),
        ),
        migrations.AlterField(
            model_name="benefit",
            name="type",
            field=models.CharField(
                blank=True,
                choices=[
                    (
                        "Percentage",
                        "Discount is a percentage off of the product's value",
                    ),
                    (
                        "Absolute",
                        "Discount is a fixed amount off of the basket's total",
                    ),
                    ("Fixed", "Discount is a fixed amount off of the product's value"),
                    ("Multibuy", "Discount is to give the cheapest product for free"),
                    (
                        "Fixed price",
                        "Get the products that meet the condition for a fixed price",
                    ),
                    (
                        "Shipping absolute",
                        "Discount is a fixed amount of the shipping cost",
                    ),
                    ("Shipping fixed price", "Get shipping for a fixed price"),
                    (
                        "Shipping percentage",
                        "Discount is a percentage off of the shipping cost",
                    ),
                ],
                max_length=128,
                verbose_name="Type",
            ),
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-01 07:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('analytics', '0002_auto_20140827_1705'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='userproductview',
            options={'ordering': ['-pk'], 'verbose_name': 'User product view', 'verbose_name_plural': 'User product views'},
        ),
        migrations.AlterModelOptions(
            name='usersearch',
            options={'ordering': ['-pk'], 'verbose_name': 'User search query', 'verbose_name_plural': 'User search queries'},
        ),
    ]



from django.apps import apps
from django.urls import include, path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class DashboardConfig(OscarDashboardConfig):
    label = "dashboard"
    name = "oscar.apps.dashboard"
    verbose_name = _("Dashboard")

    namespace = "dashboard"

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "index": (
                DashboardPermission.staff,
                DashboardPermission.partner_dashboard_access,
            ),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.index_view = get_class("dashboard.views", "IndexView")
        self.login_view = get_class("dashboard.views", "LoginView")

        self.catalogue_app = apps.get_app_config("catalogue_dashboard")
        self.reports_app = apps.get_app_config("reports_dashboard")
        self.orders_app = apps.get_app_config("orders_dashboard")
        self.users_app = apps.get_app_config("users_dashboard")
        self.pages_app = apps.get_app_config("pages_dashboard")
        self.partners_app = apps.get_app_config("partners_dashboard")
        self.offers_app = apps.get_app_config("offers_dashboard")
        self.ranges_app = apps.get_app_config("ranges_dashboard")
        self.reviews_app = apps.get_app_config("reviews_dashboard")
        self.vouchers_app = apps.get_app_config("vouchers_dashboard")
        self.comms_app = apps.get_app_config("communications_dashboard")
        self.shipping_app = apps.get_app_config("shipping_dashboard")
        self.configure_permissions()

    def get_urls(self):
        from django.contrib.auth import views as auth_views

        urls = [
            path("", self.index_view.as_view(), name="index"),
            path("catalogue/", include(self.catalogue_app.urls[0])),
            path("reports/", include(self.reports_app.urls[0])),
            path("orders/", include(self.orders_app.urls[0])),
            path("users/", include(self.users_app.urls[0])),
            path("pages/", include(self.pages_app.urls[0])),
            path("partners/", include(self.partners_app.urls[0])),
            path("offers/", include(self.offers_app.urls[0])),
            path("ranges/", include(self.ranges_app.urls[0])),
            path("reviews/", include(self.reviews_app.urls[0])),
            path("vouchers/", include(self.vouchers_app.urls[0])),
            path("comms/", include(self.comms_app.urls[0])),
            path("shipping/", include(self.shipping_app.urls[0])),
            path("login/", self.login_view.as_view(), name="login"),
            path(
                "logout/", auth_views.LogoutView.as_view(next_page="/"), name="logout"
            ),
        ]
        return self.post_process_urls(urls)



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class VoucherConfig(OscarConfig):
    label = "voucher"
    name = "oscar.apps.voucher"
    verbose_name = _("Voucher")

    def ready(self):
        # pylint: disable=unused-import
        from . import receivers



import logging

from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import login as auth_login

from oscar.apps.customer.signals import user_registered
from oscar.core.compat import get_user_model
from oscar.core.loading import get_class, get_model

User = get_user_model()
CommunicationEventType = get_model("communication", "CommunicationEventType")
Order = get_model("order", "Order")
CustomerDispatcher = get_class("customer.utils", "CustomerDispatcher")

logger = logging.getLogger("oscar.customer")


class PageTitleMixin(object):
    """
    Passes page_title and active_tab into context, which makes it quite useful
    for the accounts views.

    Dynamic page titles are possible by overriding get_page_title.
    """

    page_title = None
    active_tab = None

    # Use a method that can be overridden and customised
    def get_page_title(self):
        return self.page_title

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.setdefault("page_title", self.get_page_title())
        ctx.setdefault("active_tab", self.active_tab)
        return ctx


class RegisterUserMixin(object):
    def register_user(self, form):
        """
        Create a user instance and send a new registration email (if configured
        to).
        """
        user = form.save()

        # Raise signal robustly (we don't want exceptions to crash the request
        # handling).
        user_registered.send_robust(sender=self, request=self.request, user=user)

        if getattr(settings, "OSCAR_SEND_REGISTRATION_EMAIL", True):
            self.send_registration_email(user)

        # We have to authenticate before login
        try:
            user = authenticate(
                username=user.email, password=form.cleaned_data["password1"]
            )
        except User.MultipleObjectsReturned:
            # Handle race condition where the registration request is made
            # multiple times in quick succession.  This leads to both requests
            # passing the uniqueness check and creating users (as the first one
            # hasn't committed when the second one runs the check).  We retain
            # the first one and deactivate the dupes.
            logger.warning(
                "Multiple users with identical email address and password"
                "were found. Marking all but one as not active."
            )
            # As this section explicitly deals with the form being submitted
            # twice, this is about the only place in Oscar where we don't
            # ignore capitalisation when looking up an email address.
            # We might otherwise accidentally mark unrelated users as inactive
            users = User.objects.filter(email=user.email)
            user = users[0]
            user.backend = "oscar.apps.customer.auth_backends.EmailBackend"
            for u in users[1:]:
                u.is_active = False
                u.save()

        auth_login(self.request, user)

        return user

    def send_registration_email(self, user):
        extra_context = {"user": user, "request": self.request}
        CustomerDispatcher().send_registration_email_for_user(user, extra_context)


class AnonymousOrderMixin:
    @property
    def email(self):
        return self.request.session.get("anon_order_email")

    @property
    def order_number(self):
        return self.kwargs.get("order_number")

    @property
    def hash(self):
        return self.kwargs.get("hash")

    def get_object(self):
        try:
            order = Order.objects.get(user=None, number=self.order_number)
            if order.check_verification_hash(self.hash) and order.email == self.email:
                return order
        except (AttributeError, Order.DoesNotExist):
            return None
        return None



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class WishlistsConfig(OscarConfig):
    label = "wishlists"
    name = "oscar.apps.wishlists"
    verbose_name = _("Wishlists")

    namespace = "wishlists"

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.wishlist_view = get_class("wishlists.views", "WishListView")

    def get_urls(self):
        urls = [
            path("<str:key>/", self.wishlist_view.as_view(), name="detail"),
        ]

        return self.post_process_urls(urls)



from django.contrib import admin

from oscar.core.loading import get_model

OrderAndItemCharges = get_model("shipping", "OrderAndItemCharges")
WeightBand = get_model("shipping", "WeightBand")
WeightBased = get_model("shipping", "WeightBased")


class OrderChargesAdmin(admin.ModelAdmin):
    filter_horizontal = ("countries",)
    list_display = (
        "name",
        "description",
        "price_per_order",
        "price_per_item",
        "free_shipping_threshold",
    )


class WeightBandInline(admin.TabularInline):
    model = WeightBand


class WeightBasedAdmin(admin.ModelAdmin):
    filter_horizontal = ("countries",)
    inlines = [WeightBandInline]


admin.site.register(OrderAndItemCharges, OrderChargesAdmin)
admin.site.register(WeightBased, WeightBasedAdmin)



from oscar.core.loading import get_class

EventHandler = get_class("order.processing", "EventHandler")


class EventHandlerMixin:
    handler_class = EventHandler

    def get_handler_class(self):
        return self.handler_class

    def get_handler(self, user=None, **kwargs):
        handler_kwargs = self.get_handler_kwargs()
        handler_kwargs.update(**kwargs)
        return self.get_handler_class()(user, **handler_kwargs)

    def get_handler_kwargs(self):
        return {}



import datetime
import string

from django import forms
from django.conf import settings
from django.contrib.auth import forms as auth_forms
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.password_validation import validate_password
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ValidationError
from django.urls import resolve, Resolver404
from django.utils.crypto import get_random_string
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.apps.customer.utils import get_password_reset_url, normalise_email
from oscar.core.compat import existing_user_fields, get_user_model
from oscar.core.loading import get_class, get_model, get_profile_class
from oscar.core.utils import datetime_combine
from oscar.forms import widgets

from urllib.parse import urlparse

CustomerDispatcher = get_class("customer.utils", "CustomerDispatcher")
ProductAlert = get_model("customer", "ProductAlert")
User = get_user_model()


def generate_username():
    letters = string.ascii_letters
    allowed_chars = letters + string.digits + "_"
    uname = get_random_string(length=30, allowed_chars=allowed_chars)
    try:
        User.objects.get(username=uname)
        return generate_username()
    except User.DoesNotExist:
        return uname


class PasswordResetForm(auth_forms.PasswordResetForm):
    """
    This form takes the same structure as its parent from :py:mod:`django.contrib.auth`
    """

    def save(self, *args, domain_override=None, request=None, **kwargs):
        """
        Generates a one-use only link for resetting password and sends to the
        user.
        """
        site = get_current_site(request)
        if domain_override is not None:
            site.domain = site.name = domain_override
        for user in self.get_users(self.cleaned_data["email"]):
            self.send_password_reset_email(site, user, request)

    def send_password_reset_email(self, site, user, request=None):
        extra_context = {
            "user": user,
            "site": site,
            "reset_url": get_password_reset_url(user),
            "request": request,
        }
        CustomerDispatcher().send_password_reset_email_for_user(user, extra_context)


class EmailAuthenticationForm(AuthenticationForm):
    """
    Extends the standard django AuthenticationForm, to support 75 character
    usernames. 75 character usernames are needed to support the EmailOrUsername
    authentication backend.
    """

    username = forms.EmailField(label=_("Email address"))
    redirect_url = forms.CharField(widget=forms.HiddenInput, required=False)

    def __init__(self, host, *args, **kwargs):
        self.host = host
        super().__init__(*args, **kwargs)

    def clean_redirect_url(self):
        url = self.cleaned_data["redirect_url"].strip()
        if url and url_has_allowed_host_and_scheme(url, self.host):
            return url


class ConfirmPasswordForm(forms.Form):
    """
    Extends the standard django AuthenticationForm, to support 75 character
    usernames. 75 character usernames are needed to support the EmailOrUsername
    authentication backend.
    """

    password = forms.CharField(label=_("Password"), widget=forms.PasswordInput)

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = user

    def clean_password(self):
        password = self.cleaned_data["password"]
        if not self.user.check_password(password):
            raise forms.ValidationError(_("The entered password is not valid!"))
        return password


class EmailUserCreationForm(forms.ModelForm):
    email = forms.EmailField(label=_("Email address"))
    password1 = forms.CharField(label=_("Password"), widget=forms.PasswordInput)
    password2 = forms.CharField(label=_("Confirm password"), widget=forms.PasswordInput)
    redirect_url = forms.CharField(widget=forms.HiddenInput, required=False)

    class Meta:
        model = User
        fields = ("email",)

    def __init__(self, *args, host=None, **kwargs):
        self.host = host
        super().__init__(*args, **kwargs)

    def _post_clean(self):
        super()._post_clean()
        password = self.cleaned_data.get("password2")
        # Validate after self.instance is updated with form data
        # otherwise validators can't access email
        # see django.contrib.auth.forms.UserCreationForm
        if password:
            try:
                validate_password(password, self.instance)
            except forms.ValidationError as error:
                self.add_error("password2", error)

    def clean_email(self):
        """
        Checks for existing users with the supplied email address.
        """
        email = normalise_email(self.cleaned_data["email"])
        if User._default_manager.filter(email__iexact=email).exists():
            raise forms.ValidationError(
                _("A user with that email address already exists")
            )
        return email

    def clean_password2(self):
        password1 = self.cleaned_data.get("password1", "")
        password2 = self.cleaned_data.get("password2", "")
        if password1 != password2:
            raise forms.ValidationError(_("The two password fields didn't match."))
        return password2

    def clean_redirect_url(self):
        url = self.cleaned_data["redirect_url"].strip()
        if not url or not url_has_allowed_host_and_scheme(url, self.host):
            return settings.LOGIN_REDIRECT_URL

        try:
            parsed = urlparse(url)
            resolve(parsed.path)
        except Resolver404:
            return settings.LOGIN_REDIRECT_URL
        return url

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_password(self.cleaned_data["password1"])

        if "username" in [f.name for f in User._meta.fields]:
            user.username = generate_username()
        if commit:
            user.save()
        return user


class OrderSearchForm(forms.Form):
    date_from = forms.DateField(
        required=False,
        label=pgettext_lazy("start date", "From"),
        widget=widgets.DatePickerInput(),
    )
    date_to = forms.DateField(
        required=False,
        label=pgettext_lazy("end date", "To"),
        widget=widgets.DatePickerInput(),
    )
    order_number = forms.CharField(required=False, label=_("Order number"))

    def clean(self):
        if self.is_valid() and not any(
            [
                self.cleaned_data["date_from"],
                self.cleaned_data["date_to"],
                self.cleaned_data["order_number"],
            ]
        ):
            raise forms.ValidationError(_("At least one field is required."))
        return super().clean()

    def description(self):
        """
        Uses the form's data to build a useful description of what orders
        are listed.
        """
        if not self.is_bound or not self.is_valid():
            return _("All orders")
        else:
            date_from = self.cleaned_data["date_from"]
            date_to = self.cleaned_data["date_to"]
            order_number = self.cleaned_data["order_number"]
            return self._orders_description(date_from, date_to, order_number)

    def _orders_description(self, date_from, date_to, order_number):
        if date_from and date_to:
            if order_number:
                desc = _(
                    "Orders placed between %(date_from)s and "
                    "%(date_to)s and order number containing "
                    "%(order_number)s"
                )
            else:
                desc = _("Orders placed between %(date_from)s and %(date_to)s")
        elif date_from:
            if order_number:
                desc = _(
                    "Orders placed since %(date_from)s and "
                    "order number containing %(order_number)s"
                )
            else:
                desc = _("Orders placed since %(date_from)s")
        elif date_to:
            if order_number:
                desc = _(
                    "Orders placed until %(date_to)s and "
                    "order number containing %(order_number)s"
                )
            else:
                desc = _("Orders placed until %(date_to)s")
        elif order_number:
            desc = _("Orders with order number containing %(order_number)s")
        else:
            return None
        params = {
            "date_from": date_from,
            "date_to": date_to,
            "order_number": order_number,
        }
        return desc % params

    def get_filters(self):
        date_from = self.cleaned_data["date_from"]
        date_to = self.cleaned_data["date_to"]
        order_number = self.cleaned_data["order_number"]
        kwargs = {}
        if date_from:
            kwargs["date_placed__gte"] = datetime_combine(date_from, datetime.time.min)
        if date_to:
            kwargs["date_placed__lte"] = datetime_combine(date_to, datetime.time.max)
        if order_number:
            kwargs["number__contains"] = order_number
        return kwargs


class UserForm(forms.ModelForm):
    def __init__(self, user, *args, **kwargs):
        self.user = user
        kwargs["instance"] = user
        super().__init__(*args, **kwargs)
        if "email" in self.fields:
            self.fields["email"].required = True

    def clean_email(self):
        """
        Make sure that the email address is always unique as it is
        used instead of the username. This is necessary because the
        uniqueness of email addresses is *not* enforced on the model
        level in ``django.contrib.auth.models.User``.
        """
        email = normalise_email(self.cleaned_data["email"])
        if (
            User._default_manager.filter(email__iexact=email)
            .exclude(id=self.user.id)
            .exists()
        ):
            raise ValidationError(_("A user with this email address already exists"))
        # Save the email unaltered
        return email

    class Meta:
        model = User
        fields = existing_user_fields(["first_name", "last_name", "email"])


Profile = get_profile_class()
if Profile:

    class UserAndProfileForm(forms.ModelForm):
        def __init__(self, user, *args, **kwargs):
            try:
                instance = Profile.objects.get(user=user)
            except Profile.DoesNotExist:
                # User has no profile, try a blank one
                instance = Profile(user=user)
            kwargs["instance"] = instance

            super().__init__(*args, **kwargs)

            # Get profile field names to help with ordering later
            profile_field_names = list(self.fields.keys())

            # Get user field names (we look for core user fields first)
            core_field_names = set([f.name for f in User._meta.fields])
            user_field_names = ["email"]
            for field_name in ("first_name", "last_name"):
                if field_name in core_field_names:
                    user_field_names.append(field_name)
            user_field_names.extend(User._meta.additional_fields)

            # Store user fields so we know what to save later
            self.user_field_names = user_field_names

            # Add additional user form fields
            additional_fields = forms.fields_for_model(User, fields=user_field_names)
            self.fields.update(additional_fields)

            # Ensure email is required and initialised correctly
            self.fields["email"].required = True

            # Set initial values
            for field_name in user_field_names:
                self.fields[field_name].initial = getattr(user, field_name)

            # Ensure order of fields is email, user fields then profile fields
            self.field_order = user_field_names + profile_field_names
            self.order_fields(self.field_order)

        class Meta:
            model = Profile
            # pylint: disable=modelform-uses-exclude
            exclude = ("user",)

        def clean_email(self):
            email = normalise_email(self.cleaned_data["email"])

            users_with_email = User._default_manager.filter(
                email__iexact=email
            ).exclude(id=self.instance.user.id)
            if users_with_email.exists():
                raise ValidationError(
                    _("A user with this email address already exists")
                )
            return email

        def save(self, *args, **kwargs):
            user = self.instance.user

            # Save user also
            for field_name in self.user_field_names:
                setattr(user, field_name, self.cleaned_data[field_name])
            user.save()

            return super().save(*args, **kwargs)

    ProfileForm = UserAndProfileForm
else:
    ProfileForm = UserForm


class ProductAlertForm(forms.ModelForm):
    email = forms.EmailField(
        required=True,
        label=_("Send notification to"),
        widget=forms.TextInput(attrs={"placeholder": _("Enter your email")}),
    )

    def __init__(self, user, product, *args, **kwargs):
        self.user = user
        self.product = product
        super().__init__(*args, **kwargs)

        # Only show email field to unauthenticated users
        if user and user.is_authenticated:
            self.fields["email"].widget = forms.HiddenInput()
            self.fields["email"].required = False

    def save(self, commit=True):
        alert = super().save(commit=False)
        if self.user.is_authenticated:
            alert.user = self.user
        alert.product = self.product
        if commit:
            alert.save()
        return alert

    def clean(self):
        cleaned_data = self.cleaned_data
        email = cleaned_data.get("email")
        if email:
            try:
                ProductAlert.objects.get(
                    product=self.product,
                    email__iexact=email,
                    status=ProductAlert.ACTIVE,
                )
            except ProductAlert.DoesNotExist:
                pass
            else:
                raise forms.ValidationError(
                    _("There is already an active stock alert for %s") % email
                )

            # Check that the email address hasn't got other unconfirmed alerts.
            # If they do then we don't want to spam them with more until they
            # have confirmed or cancelled the existing alert.
            if ProductAlert.objects.filter(
                email__iexact=email, status=ProductAlert.UNCONFIRMED
            ).count():
                raise forms.ValidationError(
                    _(
                        "%s has been sent a confirmation email for another product "
                        "alert on this site. Please confirm or cancel that request "
                        "before signing up for more alerts."
                    )
                    % email
                )
        elif self.user.is_authenticated:
            try:
                ProductAlert.objects.get(
                    product=self.product, user=self.user, status=ProductAlert.ACTIVE
                )
            except ProductAlert.DoesNotExist:
                pass
            else:
                raise forms.ValidationError(
                    _("You already have an active alert for this product")
                )
        return cleaned_data

    class Meta:
        model = ProductAlert
        fields = ["email"]


class AnonymousOrderForm(forms.Form):
    email = forms.EmailField(label=_("Email address"))



from oscar.core.loading import get_model

Category = get_model("catalogue", "category")


def create_from_sequence(bits):
    """
    Create categories from an iterable
    """
    if len(bits) == 1:
        # Get or create root node
        name = bits[0]
        try:
            # Category names should be unique at the depth=1
            root = Category.objects.get(depth=1, name=name)
        except Category.DoesNotExist:
            root = Category.add_root(name=name)
        except Category.MultipleObjectsReturned:
            raise ValueError(
                ("There are more than one categories with name %s at depth=1") % name
            )
        return [root]
    else:
        parents = create_from_sequence(bits[:-1])
        parent, name = parents[-1], bits[-1]
        try:
            child = parent.get_children().get(name=name)
        except Category.DoesNotExist:
            child = parent.add_child(name=name)
        except Category.MultipleObjectsReturned:
            raise ValueError(
                (
                    "There are more than one categories with name "
                    "%s which are children of %s"
                )
                % (name, parent)
            )
        parents.append(child)
        return parents


def create_from_breadcrumbs(breadcrumb_str, separator=">"):
    """
    Create categories from a breadcrumb string
    """
    category_names = [x.strip() for x in breadcrumb_str.split(separator)]
    categories = create_from_sequence(category_names)
    return categories[-1]



# pylint: disable=attribute-defined-outside-init
from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DeleteView, DetailView, FormView, ListView, UpdateView
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import FormMixin
from django_tables2 import SingleTableView

from oscar.core.compat import get_user_model
from oscar.core.loading import get_class, get_classes, get_model
from oscar.views.generic import BulkEditMixin

UserSearchForm, ProductAlertSearchForm, ProductAlertUpdateForm = get_classes(
    "dashboard.users.forms",
    ("UserSearchForm", "ProductAlertSearchForm", "ProductAlertUpdateForm"),
)
PasswordResetForm = get_class("customer.forms", "PasswordResetForm")
UserTable = get_class("dashboard.users.tables", "UserTable")
ProductAlert = get_model("customer", "ProductAlert")
User = get_user_model()


class IndexView(BulkEditMixin, FormMixin, SingleTableView):
    template_name = "oscar/dashboard/users/index.html"
    model = User
    actions = (
        "make_active",
        "make_inactive",
    )
    form_class = UserSearchForm
    table_class = UserTable
    context_table_name = "users"
    desc_template = _("%(main_filter)s %(email_filter)s %(name_filter)s")
    description = ""

    def dispatch(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        self.form = self.get_form(form_class)
        return super().dispatch(request, *args, **kwargs)

    def get_table_pagination(self, table):
        return dict(per_page=settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE)

    def get_form_kwargs(self):
        """
        Only bind search form if it was submitted.
        """
        kwargs = super().get_form_kwargs()

        if "search" in self.request.GET:
            kwargs.update(
                {
                    "data": self.request.GET,
                }
            )

        return kwargs

    def get_queryset(self):
        queryset = self.model.objects.select_related("userrecord").order_by(
            "-date_joined"
        )
        return self.apply_search(queryset)

    def apply_search(self, queryset):
        # Set initial queryset description, used for template context
        self.desc_ctx = {
            "main_filter": _("All users"),
            "email_filter": "",
            "name_filter": "",
        }
        if self.form.is_valid():
            return self.apply_search_filters(queryset, self.form.cleaned_data)
        else:
            return queryset

    def apply_search_filters(self, queryset, data):
        """
        Function is split out to allow customisation with little boilerplate.
        """
        if data["email"]:
            email = data["email"]
            queryset = queryset.filter(email__istartswith=email)
            self.desc_ctx["email_filter"] = _(" with email matching '%s'") % email
        if data["name"]:
            # If the value is two words, then assume they are first name and
            # last name
            parts = data["name"].split()
            # always true filter
            condition = Q()
            for part in parts:
                condition &= Q(first_name__icontains=part) | Q(
                    last_name__icontains=part
                )
            queryset = queryset.filter(condition).distinct()
            self.desc_ctx["name_filter"] = _(" with name matching '%s'") % data["name"]

        return queryset

    def get_table(self, **kwargs):
        table = super().get_table(**kwargs)
        table.caption = self.desc_template % self.desc_ctx
        return table

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form"] = self.form
        return context

    def make_inactive(self, request, users):
        return self._change_users_active_status(users, False)

    def make_active(self, request, users):
        return self._change_users_active_status(users, True)

    def _change_users_active_status(self, users, value):
        for user in users:
            if not user.is_superuser:
                user.is_active = value
                user.save()
        messages.info(self.request, _("Users' status successfully changed"))
        return redirect("dashboard:users-index")


class UserDetailView(DetailView):
    template_name = "oscar/dashboard/users/detail.html"
    model = User
    context_object_name = "customer"

    def get_queryset(self):
        queryset = self.model.objects.prefetch_related(
            "orders__lines", "orders__surcharges"
        )
        return queryset


class PasswordResetView(SingleObjectMixin, FormView):
    form_class = PasswordResetForm
    http_method_names = ["post"]
    model = User

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["data"] = {"email": self.object.email}
        return kwargs

    def form_valid(self, form):
        # The PasswordResetForm's save method sends the reset email
        form.save(request=self.request)
        return super().form_valid(form)

    def get_success_url(self):
        messages.success(self.request, _("A password reset email has been sent"))
        return reverse("dashboard:user-detail", kwargs={"pk": self.object.id})


class ProductAlertListView(ListView):
    model = ProductAlert
    form_class = ProductAlertSearchForm
    context_object_name = "alerts"
    template_name = "oscar/dashboard/users/alerts/list.html"
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE
    base_description = _("All Alerts")
    description = ""

    def get_queryset(self):
        queryset = self.model.objects.all().order_by("-date_created")
        self.description = self.base_description

        self.form = self.form_class(self.request.GET)
        if not self.form.is_valid():
            return queryset

        data = self.form.cleaned_data

        if data["status"]:
            queryset = queryset.filter(status=data["status"])
            self.description += _(" with status matching '%s'") % data["status"]

        if data["name"]:
            # If the value is two words, then assume they are first name and
            # last name
            parts = data["name"].split()
            if len(parts) >= 2:
                queryset = queryset.filter(
                    user__first_name__istartswith=parts[0],
                    user__last_name__istartswith=parts[1],
                ).distinct()
            else:
                queryset = queryset.filter(
                    Q(user__first_name__istartswith=parts[0])
                    | Q(user__last_name__istartswith=parts[-1])
                ).distinct()
            self.description += _(" with customer name matching '%s'") % data["name"]

        if data["email"]:
            queryset = queryset.filter(
                Q(user__email__icontains=data["email"])
                | Q(email__icontains=data["email"])
            )
            self.description += _(" with customer email matching '%s'") % data["email"]

        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form"] = self.form
        context["queryset_description"] = self.description
        return context


class ProductAlertUpdateView(UpdateView):
    template_name = "oscar/dashboard/users/alerts/update.html"
    model = ProductAlert
    form_class = ProductAlertUpdateForm
    context_object_name = "alert"

    def get_success_url(self):
        messages.success(self.request, _("Product alert saved"))
        return reverse("dashboard:user-alert-list")


class ProductAlertDeleteView(DeleteView):
    model = ProductAlert
    template_name = "oscar/dashboard/users/alerts/delete.html"
    context_object_name = "alert"

    def get_success_url(self):
        messages.warning(self.request, _("Product alert deleted"))
        return reverse("dashboard:user-alert-list")



from django.conf import settings

from haystack.generic_views import FacetedSearchView as BaseFacetedSearchView

from oscar.core.loading import get_class

FacetMunger = get_class("search.facets", "FacetMunger")
base_sqs = get_class("search.facets", "base_sqs")


class BaseSearchView(BaseFacetedSearchView):
    facet_fields = settings.OSCAR_SEARCH_FACETS["fields"].keys()
    paginate_by = settings.OSCAR_PRODUCTS_PER_PAGE

    def get_queryset(self):
        return base_sqs()

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)

        form = context[self.form_name]

        # Show suggestion no matter what.  Haystack 2.1 only shows a suggestion
        # if there are some results, which seems a bit weird to me.
        if self.queryset.query.backend.include_spelling:
            # Note, this triggers an extra call to the search backend
            suggestion = form.get_suggestion()
            if suggestion != context["query"]:
                context["suggestion"] = suggestion

        # Convert facet data into a more useful data structure
        if "fields" in context["facets"]:
            munger = FacetMunger(
                self.request.get_full_path(),
                form.selected_multi_facets,
                self.queryset.facet_counts(),
                query_type=type(self.queryset.query),
            )
            context["facet_data"] = munger.facet_data()
            has_facets = any(
                [len(data["results"]) for data in context["facet_data"].values()]
            )
            context["has_facets"] = has_facets

        context["selected_facets"] = form.selected_facets
        context[self.page_kwarg] = context["page_obj"]

        return context



import logging
import os

from django.conf import settings


def get_file_logger(filename, level=logging.INFO):
    """
    Return a logger that uses a file handler for the given filename but in the
    LOG_ROOT.
    """
    filepath = os.path.join(settings.LOG_ROOT, filename)
    handler = logging.FileHandler(filepath)
    handler.setLevel(level)
    logger = logging.getLogger(filepath)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger



from django.db.models import Q


class PartnerProductFilterMixin:
    def filter_queryset(self, queryset):
        """
        Restrict the queryset to products the given user has access to.
        A staff user is allowed to access all Products.
        A non-staff user is only allowed access to a product if they are in at
        least one stock record's partner user list.
        """
        user = self.request.user
        if user.is_staff:
            return queryset

        return queryset.filter(
            Q(children__stockrecords__partner__users__pk=user.pk)
            | Q(stockrecords__partner__users__pk=user.pk)
        ).distinct()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.core.utils


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='currency',
            field=models.CharField(default=oscar.core.utils.get_default_currency, max_length=12, verbose_name='Currency'),
        ),
    ]



# pylint: disable=attribute-defined-outside-init
from django.contrib import messages
from django.contrib.auth.models import Permission
from django.shortcuts import get_object_or_404, redirect
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views import generic

from oscar.apps.customer.utils import normalise_email
from oscar.core.compat import get_user_model
from oscar.core.loading import get_classes, get_model
from oscar.views import sort_queryset

User = get_user_model()
Partner = get_model("partner", "Partner")
(
    PartnerSearchForm,
    PartnerCreateForm,
    PartnerAddressForm,
    NewUserForm,
    UserEmailForm,
    ExistingUserForm,
) = get_classes(
    "dashboard.partners.forms",
    [
        "PartnerSearchForm",
        "PartnerCreateForm",
        "PartnerAddressForm",
        "NewUserForm",
        "UserEmailForm",
        "ExistingUserForm",
    ],
)


class PartnerListView(generic.ListView):
    model = Partner
    context_object_name = "partners"
    template_name = "oscar/dashboard/partners/partner_list.html"
    form_class = PartnerSearchForm

    def get_queryset(self):
        qs = self.model._default_manager.all()
        qs = sort_queryset(qs, self.request, ["name"])

        self.description = _("All partners")

        # We track whether the queryset is filtered to determine whether we
        # show the search form 'reset' button.
        self.is_filtered = False
        self.form = self.form_class(self.request.GET)
        if not self.form.is_valid():
            return qs

        data = self.form.cleaned_data

        if data["name"]:
            qs = qs.filter(name__icontains=data["name"])
            self.description = _("Partners matching '%s'") % data["name"]
            self.is_filtered = True

        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["queryset_description"] = self.description
        ctx["form"] = self.form
        ctx["is_filtered"] = self.is_filtered
        return ctx


class PartnerCreateView(generic.CreateView):
    model = Partner
    template_name = "oscar/dashboard/partners/partner_form.html"
    form_class = PartnerCreateForm
    success_url = reverse_lazy("dashboard:partner-list")

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Create new partner")
        return ctx

    def get_success_url(self):
        messages.success(
            self.request, _("Partner '%s' was created successfully.") % self.object.name
        )
        return reverse("dashboard:partner-list")


class PartnerManageView(generic.UpdateView):
    """
    This multi-purpose view renders out a form to edit the partner's details,
    the associated address and a list of all associated users.
    """

    template_name = "oscar/dashboard/partners/partner_manage.html"
    form_class = PartnerAddressForm
    success_url = reverse_lazy("dashboard:partner-list")

    def get_object(self, queryset=None):
        self.partner = get_object_or_404(Partner, pk=self.kwargs["pk"])
        address = self.partner.primary_address
        if address is None:
            address = self.partner.addresses.model(partner=self.partner)
        return address

    def get_initial(self):
        return {"name": self.partner.name}

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["partner"] = self.partner
        ctx["title"] = self.partner.name
        ctx["users"] = self.partner.users.all()
        return ctx

    def form_valid(self, form):
        messages.success(
            self.request,
            _("Partner '%s' was updated successfully.") % self.partner.name,
        )
        self.partner.name = form.cleaned_data["name"]
        self.partner.save()
        return super().form_valid(form)


class PartnerDeleteView(generic.DeleteView):
    model = Partner
    template_name = "oscar/dashboard/partners/partner_delete.html"

    def get_success_url(self):
        messages.success(
            self.request, _("Partner '%s' was deleted successfully.") % self.object.name
        )
        return reverse("dashboard:partner-list")


# =============
# Partner users
# =============


class PartnerUserCreateView(generic.CreateView):
    model = User
    template_name = "oscar/dashboard/partners/partner_user_form.html"
    form_class = NewUserForm

    def dispatch(self, request, *args, **kwargs):
        self.partner = get_object_or_404(Partner, pk=kwargs.get("partner_pk", None))
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["partner"] = self.partner
        ctx["title"] = _("Create user")
        return ctx

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["partner"] = self.partner
        return kwargs

    def get_success_url(self):
        name = self.object.get_full_name() or self.object.email
        messages.success(self.request, _("User '%s' was created successfully.") % name)
        return reverse("dashboard:partner-list")


class PartnerUserSelectView(generic.ListView):
    template_name = "oscar/dashboard/partners/partner_user_select.html"
    form_class = UserEmailForm
    context_object_name = "users"

    def dispatch(self, request, *args, **kwargs):
        self.partner = get_object_or_404(Partner, pk=kwargs.get("partner_pk", None))
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        data = None
        if "email" in request.GET:
            data = request.GET
        self.form = self.form_class(data)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["partner"] = self.partner
        ctx["form"] = self.form
        return ctx

    def get_queryset(self):
        if self.form.is_valid():
            email = normalise_email(self.form.cleaned_data["email"])
            return User.objects.filter(email__icontains=email)
        else:
            return User.objects.none()


class PartnerUserLinkView(generic.View):
    def get(self, request, user_pk, partner_pk):
        # need to allow GET to make Undo link in PartnerUserUnlinkView work
        return self.post(request, user_pk, partner_pk)

    def post(self, request, user_pk, partner_pk):
        user = get_object_or_404(User, pk=user_pk)
        name = user.get_full_name() or user.email
        partner = get_object_or_404(Partner, pk=partner_pk)
        if self.link_user(user, partner):
            messages.success(
                request,
                _("User '%(name)s' was linked to '%(partner_name)s'")
                % {"name": name, "partner_name": partner.name},
            )
        else:
            messages.info(
                request,
                _("User '%(name)s' is already linked to '%(partner_name)s'")
                % {"name": name, "partner_name": partner.name},
            )
        return redirect("dashboard:partner-manage", pk=partner_pk)

    def link_user(self, user, partner):
        """
        Links a user to a partner, and adds the dashboard permission if needed.

        Returns False if the user was linked already; True otherwise.
        """
        if partner.users.filter(pk=user.pk).exists():
            return False
        partner.users.add(user)
        if not user.is_staff:
            dashboard_access_perm = Permission.objects.get(
                codename="dashboard_access", content_type__app_label="partner"
            )
            user.user_permissions.add(dashboard_access_perm)
        return True


class PartnerUserUnlinkView(generic.View):
    def unlink_user(self, user, partner):
        """
        Unlinks a user from a partner, and removes the dashboard permission
        if they are not linked to any other partners.

        Returns False if the user was not linked to the partner; True
        otherwise.
        """
        if not partner.users.filter(pk=user.pk).exists():
            return False
        partner.users.remove(user)
        if not user.is_staff and not user.partners.exists():
            dashboard_access_perm = Permission.objects.get(
                codename="dashboard_access", content_type__app_label="partner"
            )
            user.user_permissions.remove(dashboard_access_perm)
        return True

    def post(self, request, user_pk, partner_pk):
        user = get_object_or_404(User, pk=user_pk)
        name = user.get_full_name() or user.email
        partner = get_object_or_404(Partner, pk=partner_pk)
        if self.unlink_user(user, partner):
            msg = render_to_string(
                "oscar/dashboard/partners/messages/user_unlinked.html",
                {
                    "user_name": name,
                    "partner_name": partner.name,
                    "user_pk": user_pk,
                    "partner_pk": partner_pk,
                },
            )
            messages.success(self.request, msg, extra_tags="safe noicon")
        else:
            messages.error(
                request,
                _("User '%(name)s' is not linked to '%(partner_name)s'")
                % {"name": name, "partner_name": partner.name},
            )
        return redirect("dashboard:partner-manage", pk=partner_pk)


# =====
# Users
# =====


class PartnerUserUpdateView(generic.UpdateView):
    template_name = "oscar/dashboard/partners/partner_user_form.html"
    form_class = ExistingUserForm

    def get_object(self, queryset=None):
        self.partner = get_object_or_404(Partner, pk=self.kwargs["partner_pk"])
        return get_object_or_404(
            User, pk=self.kwargs["user_pk"], partners__pk=self.kwargs["partner_pk"]
        )

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        name = self.object.get_full_name() or self.object.email
        ctx["partner"] = self.partner
        ctx["title"] = _("Edit user '%s'") % name
        return ctx

    def get_success_url(self):
        name = self.object.get_full_name() or self.object.email
        messages.success(self.request, _("User '%s' was updated successfully.") % name)
        return reverse("dashboard:partner-list")



from collections import Counter

from django.core.management.base import BaseCommand

from oscar.core.compat import get_user_model

User = get_user_model()


class Command(BaseCommand):
    help = (
        "Finds email addresses that are used by more than one user. "
        "Casing is ignored."
    )

    def handle(self, *args, **options):
        emails = User.objects.values_list("email", flat=True)
        emails = map(lambda x: x.lower(), emails)
        duplicates = sorted(
            [
                (email, count)
                for email, count in Counter(emails).most_common()
                if count > 1
            ]
        )
        if duplicates:
            for email, __ in duplicates:
                users = User.objects.filter(email__iexact=email)
                user_strings = [
                    "{} (#{})".format(user.get_username(), user.pk) for user in users
                ]
                print(
                    "{email} is assigned to: {pk_list}".format(
                        email=email, pk_list=", ".join(user_strings)
                    )
                )
        else:
            print("No duplicate email addresses found!")



from oscar.core.loading import get_class

SearchForm = get_class("search.forms", "SearchForm")


def search_form(request):
    """
    Ensure that the search form is available site wide
    """
    return {"search_form": SearchForm(request.GET)}



from django.db import models, router
from django.db.models import F, signals
from django.db.models.functions import Coalesce, Least
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.apps.partner.exceptions import InvalidStockAdjustment
from oscar.core.compat import AUTH_USER_MODEL
from oscar.core.utils import get_default_currency
from oscar.models.fields import AutoSlugField


class AbstractPartner(models.Model):
    """
    A fulfilment partner. An individual or company who can fulfil products.
    E.g. for physical goods, somebody with a warehouse and means of delivery.

    Creating one or more instances of the Partner model is a required step in
    setting up an Oscar deployment. Many Oscar deployments will only have one
    fulfilment partner.
    """

    code = AutoSlugField(
        _("Code"), max_length=128, unique=True, db_index=True, populate_from="name"
    )
    name = models.CharField(
        pgettext_lazy("Partner's name", "Name"),
        max_length=128,
        blank=True,
        db_index=True,
    )

    #: A partner can have users assigned to it. This is used
    #: for access modelling in the permission-based dashboard
    users = models.ManyToManyField(
        AUTH_USER_MODEL, related_name="partners", blank=True, verbose_name=_("Users")
    )

    @property
    def display_name(self):
        return self.name or self.code

    @property
    def primary_address(self):
        """
        Returns a partners primary address. Usually that will be the
        headquarters or similar.

        This is a rudimentary implementation that raises an error if there's
        more than one address. If you actually want to support multiple
        addresses, you will likely need to extend PartnerAddress to have some
        field or flag to base your decision on.
        """
        addresses = self.addresses.all()
        if len(addresses) == 0:  # intentionally using len() to save queries
            return None
        elif len(addresses) == 1:
            return addresses[0]
        else:
            raise NotImplementedError(
                "Oscar's default implementation of primary_address only "
                "supports one PartnerAddress.  You need to override the "
                "primary_address to look up the right address"
            )

    # pylint: disable=unused-argument
    def get_address_for_stockrecord(self, stockrecord):
        """
        Stock might be coming from different warehouses. Overriding this
        function allows selecting the correct PartnerAddress for the record.
        That can be useful when determining tax.
        """
        return self.primary_address

    class Meta:
        abstract = True
        app_label = "partner"
        ordering = ("name", "code")
        permissions = (("dashboard_access", "Can access dashboard"),)
        verbose_name = _("Fulfillment partner")
        verbose_name_plural = _("Fulfillment partners")

    def __str__(self):
        return self.display_name


class AbstractStockRecord(models.Model):
    """
    A stock record.

    This records information about a product from a fulfilment partner, such as
    their SKU, the number they have in stock and price information.

    Stockrecords are used by 'strategies' to determine availability and pricing
    information for the customer.
    """

    product = models.ForeignKey(
        "catalogue.Product",
        on_delete=models.CASCADE,
        related_name="stockrecords",
        verbose_name=_("Product"),
    )
    partner = models.ForeignKey(
        "partner.Partner",
        on_delete=models.CASCADE,
        verbose_name=_("Partner"),
        related_name="stockrecords",
    )

    #: The fulfilment partner will often have their own SKU for a product,
    #: which we store here.  This will sometimes be the same the product's UPC
    #: but not always.  It should be unique per partner.
    #: See also http://en.wikipedia.org/wiki/Stock-keeping_unit
    partner_sku = models.CharField(_("Partner SKU"), max_length=128)

    # Price info:
    price_currency = models.CharField(
        _("Currency"), max_length=12, default=get_default_currency
    )

    # This is the base price for calculations - whether this is inclusive or exclusive of
    # tax depends on your implementation, as this is highly domain-specific.
    # It is nullable because some items don't have a fixed
    # price but require a runtime calculation (possibly from an external service).
    price = models.DecimalField(
        _("Price"), decimal_places=2, max_digits=12, blank=True, null=True
    )

    #: Number of items in stock
    num_in_stock = models.PositiveIntegerField(
        _("Number in stock"), blank=True, null=True
    )

    #: The amount of stock allocated to orders but not fed back to the master
    #: stock system.  A typical stock update process will set the
    #: :py:attr:`.num_in_stock` variable to a new value and reset
    #: :py:attr:`.num_allocated` to zero.
    num_allocated = models.IntegerField(_("Number allocated"), blank=True, null=True)

    #: Threshold for low-stock alerts.  When stock goes beneath this threshold,
    #: an alert is triggered so warehouse managers can order more.
    low_stock_threshold = models.PositiveIntegerField(
        _("Low Stock Threshold"), blank=True, null=True
    )

    # Date information
    date_created = models.DateTimeField(_("Date created"), auto_now_add=True)
    date_updated = models.DateTimeField(_("Date updated"), auto_now=True, db_index=True)

    def __str__(self):
        msg = "Partner: %s, product: %s" % (
            self.partner.display_name,
            self.product,
        )
        if self.partner_sku:
            msg = "%s (%s)" % (msg, self.partner_sku)
        return msg

    class Meta:
        abstract = True
        app_label = "partner"
        unique_together = ("partner", "partner_sku")
        verbose_name = _("Stock record")
        verbose_name_plural = _("Stock records")

    @property
    def net_stock_level(self):
        """
        The effective number in stock (e.g. available to buy).

        This is correct property to show the customer, not the
        :py:attr:`.num_in_stock` field as that doesn't account for allocations.
        This can be negative in some unusual circumstances
        """
        if self.num_in_stock is None:
            return 0
        if self.num_allocated is None:
            return self.num_in_stock
        return self.num_in_stock - self.num_allocated

    @cached_property
    def can_track_allocations(self):
        """Return True if the Product is set for stock tracking."""
        return self.product.get_product_class().track_stock

    # 2-stage stock management model

    def allocate(self, quantity):
        """
        Record a stock allocation.

        This normally happens when a product is bought at checkout.  When the
        product is actually shipped, then we 'consume' the allocation.

        """
        # Doesn't make sense to allocate if stock tracking is off.
        if not self.can_track_allocations:
            return

        # Send the pre-save signal
        self.pre_save_signal()

        # Atomic update
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=(Coalesce(F("num_allocated"), 0) + quantity)
            )
        )

        # Make sure the current object is up-to-date
        self.refresh_from_db(fields=["num_allocated"])

        # Send the post-save signal
        self.post_save_signal()

    allocate.alters_data = True

    def is_allocation_consumption_possible(self, quantity):
        """
        Test if a proposed stock consumption is permitted
        """
        return quantity <= min(self.num_allocated or 0, self.num_in_stock)

    def consume_allocation(self, quantity):
        """
        Consume a previous allocation

        This is used when an item is shipped.  We remove the original
        allocation and adjust the number in stock accordingly
        """
        if not self.can_track_allocations:
            return
        if not self.is_allocation_consumption_possible(quantity):
            raise InvalidStockAdjustment(_("Invalid stock consumption request"))

        # send the pre save signal
        self.pre_save_signal()

        # Atomically consume allocations and stock
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=(Coalesce(F("num_allocated"), 0) - quantity),
                num_in_stock=(Coalesce(F("num_in_stock"), 0) - quantity),
            )
        )

        # Make sure current object is up-to-date
        self.refresh_from_db(fields=["num_allocated", "num_in_stock"])

        # Send the post-save signal
        self.post_save_signal()

    consume_allocation.alters_data = True

    def cancel_allocation(self, quantity):
        if not self.can_track_allocations:
            return

        # send the pre save signal
        self.pre_save_signal()

        # Atomically consume allocations
        (
            self.__class__.objects.filter(pk=self.pk).update(
                num_allocated=Coalesce(F("num_allocated"), 0)
                - Least(Coalesce(F("num_allocated"), 0), quantity),
            )
        )

        # Make sure current object is up-to-date
        self.refresh_from_db(fields=["num_allocated"])

        # Send the post-save signal
        self.post_save_signal()

    cancel_allocation.alters_data = True

    def pre_save_signal(self):
        signals.pre_save.send(
            sender=self.__class__,
            instance=self,
            created=False,
            raw=False,
            using=router.db_for_write(self.__class__, instance=self),
        )

    def post_save_signal(self):
        signals.post_save.send(
            sender=self.__class__,
            instance=self,
            created=False,
            raw=False,
            using=router.db_for_write(self.__class__, instance=self),
        )

    @property
    def is_below_threshold(self):
        if self.low_stock_threshold is None:
            return False
        return self.net_stock_level < self.low_stock_threshold


class AbstractStockAlert(models.Model):
    """
    A stock alert. E.g. used to notify users when a product is 'back in stock'.
    """

    stockrecord = models.ForeignKey(
        "partner.StockRecord",
        on_delete=models.CASCADE,
        related_name="alerts",
        verbose_name=_("Stock Record"),
    )
    threshold = models.PositiveIntegerField(_("Threshold"))
    OPEN, CLOSED = "Open", "Closed"
    status_choices = (
        (OPEN, _("Open")),
        (CLOSED, _("Closed")),
    )
    status = models.CharField(
        _("Status"), max_length=128, default=OPEN, choices=status_choices
    )
    date_created = models.DateTimeField(
        _("Date Created"), auto_now_add=True, db_index=True
    )
    date_closed = models.DateTimeField(_("Date Closed"), blank=True, null=True)

    def close(self):
        self.status = self.CLOSED
        self.date_closed = now()
        self.save()

    close.alters_data = True

    def __str__(self):
        return _('<stockalert for "%(stock)s" status %(status)s>') % {
            "stock": self.stockrecord,
            "status": self.status,
        }

    class Meta:
        abstract = True
        app_label = "partner"
        ordering = ("-date_created",)
        verbose_name = _("Stock alert")
        verbose_name_plural = _("Stock alerts")



from decimal import Decimal as D

from django.utils.translation import gettext_lazy as _

from oscar.core import prices
from oscar.core.utils import round_half_up_four_dec


class Base(object):
    """
    Shipping method interface class

    This is the superclass to the classes in this module. This allows
    using all shipping methods interchangeably (aka polymorphism).

    The interface is all properties.
    """

    #: Used to store this method in the session.  Each shipping method should
    #:  have a unique code.
    code = "__default__"

    #: The name of the shipping method, shown to the customer during checkout
    name = "Default shipping"

    #: A more detailed description of the shipping method shown to the customer
    #:  during checkout.  Can contain HTML.
    description = ""

    #: Whether the charge includes a discount
    is_discounted = False

    def calculate(self, basket):
        """
        Return the shipping charge for the given basket
        """
        raise NotImplementedError

    # pylint: disable=unused-argument
    def discount(self, basket):
        """
        Return the discount on the standard shipping charge
        """
        # The regular shipping methods don't add a default discount.
        # For offers and vouchers, the discount will be provided
        # by a wrapper that Repository.apply_shipping_offer() adds.
        return D("0.00")


class Free(Base):
    """
    This shipping method specifies that shipping is free.
    """

    code = "free-shipping"
    name = _("Free shipping")

    def calculate(self, basket):
        # If the charge is free then tax must be free (musn't it?) and so we
        # immediately set the tax to zero
        return prices.Price(currency=basket.currency, excl_tax=D("0.00"), tax=D("0.00"))


class NoShippingRequired(Free):
    """
    This is a special shipping method that indicates that no shipping is
    actually required (e.g. for digital goods).
    """

    code = "no-shipping-required"
    name = _("No shipping required")


class FixedPrice(Base):
    """
    This shipping method indicates that shipping costs a fixed price and
    requires no special calculation.
    """

    code = "fixed-price-shipping"
    name = _("Fixed price shipping")

    # Charges can be either declared by subclassing and overriding the
    # class attributes or by passing them to the constructor
    charge_excl_tax = None
    charge_incl_tax = None

    def __init__(self, charge_excl_tax=None, charge_incl_tax=None):
        if charge_excl_tax is not None:
            self.charge_excl_tax = charge_excl_tax
        if charge_incl_tax is not None:
            self.charge_incl_tax = charge_incl_tax

    def calculate(self, basket):
        return prices.Price(
            currency=basket.currency,
            excl_tax=self.charge_excl_tax,
            incl_tax=self.charge_incl_tax,
        )


# pylint: disable=abstract-method
class OfferDiscount(Base):
    """
    Wrapper class that applies a discount to an existing shipping
    method's charges.
    """

    is_discounted = True

    def __init__(self, method, offer):
        self.method = method
        self.offer = offer

    # Forwarded properties

    @property
    def code(self):
        """
        Returns the :py:attr:`code <oscar.apps.shipping.methods.Base.code>` of the wrapped shipping method.
        """
        return self.method.code

    @property
    def name(self):
        """
        Returns the :py:attr:`name <oscar.apps.shipping.methods.Base.name>` of the wrapped shipping method.
        """
        return self.method.name

    @property
    def discount_name(self):
        """
        Returns the :py:attr:`name <oscar.apps.offer.abstract_models.BaseOfferMixin.name>` of the applied Offer.
        """
        return self.offer.name

    @property
    def description(self):
        """
        Returns the :py:attr:`description <.Base.description>` of the wrapped shipping method.
        """
        return self.method.description

    def calculate_excl_discount(self, basket):
        """
        Returns the shipping charge for the given basket without
        discount applied.
        """
        return self.method.calculate(basket)


class TaxExclusiveOfferDiscount(OfferDiscount):
    """
    Wrapper class which extends OfferDiscount to be exclusive of tax.
    """

    def calculate(self, basket):
        base_charge = self.method.calculate(basket)
        discount = self.offer.shipping_discount(
            base_charge.excl_tax, base_charge.currency
        )
        excl_tax = base_charge.excl_tax - discount
        return prices.Price(
            currency=base_charge.currency,
            excl_tax=excl_tax,
            tax_code=base_charge.tax_code,
        )

    def discount(self, basket):
        base_charge = self.method.calculate(basket)
        return self.offer.shipping_discount(base_charge.excl_tax, base_charge.currency)


class TaxInclusiveOfferDiscount(OfferDiscount):
    """
    Wrapper class which extends OfferDiscount to be inclusive of tax.
    """

    def calculate(self, basket):
        base_charge = self.method.calculate(basket)
        discount = self.offer.shipping_discount(
            base_charge.incl_tax, base_charge.currency
        )
        incl_tax = base_charge.incl_tax - discount
        excl_tax = self.calculate_excl_tax(base_charge, incl_tax)
        return prices.Price(
            currency=base_charge.currency,
            excl_tax=excl_tax,
            incl_tax=incl_tax,
            tax_code=base_charge.tax_code,
        )

    def calculate_excl_tax(self, base_charge, incl_tax):
        """
        Return the charge excluding tax (but including discount).
        """
        if incl_tax == D("0.00"):
            return D("0.00")
        # We assume we can linearly scale down the excl tax price before
        # discount.
        excl_tax = base_charge.excl_tax * (incl_tax / base_charge.incl_tax)
        return round_half_up_four_dec(excl_tax)

    def discount(self, basket):
        base_charge = self.method.calculate(basket)
        return self.offer.shipping_discount(base_charge.incl_tax, base_charge.currency)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from oscar.core.loading import get_class

Node = get_class("dashboard.nav", "Node")


def get_nodes(user):
    """
    Return the visible navigation nodes for the passed user
    """
    all_nodes = create_menu(settings.OSCAR_DASHBOARD_NAVIGATION)
    visible_nodes = []
    for node in all_nodes:
        filtered_node = node.filter(user)
        # don't append headings without children
        if filtered_node and (
            filtered_node.has_children() or not filtered_node.is_heading
        ):
            visible_nodes.append(filtered_node)
    return visible_nodes


def create_menu(menu_items, parent=None):
    """
    Create the navigation nodes based on a passed list of dicts
    """
    nodes = []
    default_fn = import_string(settings.OSCAR_DASHBOARD_DEFAULT_ACCESS_FUNCTION)
    for menu_dict in menu_items:
        try:
            label = menu_dict["label"]
        except KeyError:
            raise ImproperlyConfigured("No label specified for menu item in dashboard")

        children = menu_dict.get("children", [])
        if children:
            node = Node(
                label=label,
                icon=menu_dict.get("icon", None),
                access_fn=menu_dict.get("access_fn", default_fn),
            )
            create_menu(children, parent=node)
        else:
            node = Node(
                label=label,
                icon=menu_dict.get("icon", None),
                url_name=menu_dict.get("url_name", None),
                url_kwargs=menu_dict.get("url_kwargs", None),
                url_args=menu_dict.get("url_args", None),
                access_fn=menu_dict.get("access_fn", default_fn),
            )
        if parent is None:
            nodes.append(node)
        else:
            parent.add_child(node)
    return nodes



import re

from django.conf import settings


def strip_language_code(request):
    """
    When using Django's i18n_patterns, we need a language-neutral variant of
    the current URL to be able to use set_language to change languages.
    This naive approach strips the language code from the beginning of the URL
    and will likely fail if using translated URLs.
    """
    path = request.path
    if settings.USE_I18N and hasattr(request, "LANGUAGE_CODE"):
        return re.sub("^/%s/" % request.LANGUAGE_CODE, "/", path)
    return path


def metadata(request):
    """
    Add some generally useful metadata to the template context
    """
    return {
        "shop_name": settings.OSCAR_SHOP_NAME,
        "shop_tagline": settings.OSCAR_SHOP_TAGLINE,
        "homepage_url": settings.OSCAR_HOMEPAGE,
        "language_neutral_url_path": strip_language_code(request),
        # Fallback to old settings name for backwards compatibility
        "google_analytics_id": (
            getattr(settings, "OSCAR_GOOGLE_ANALYTICS_ID", None)
            or getattr(settings, "GOOGLE_ANALYTICS_ID", None)
        ),
    }



# Generated by Django 2.0.7 on 2018-11-15 19:53

from decimal import Decimal
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('shipping', '0002_auto_20150604_1450'),
    ]

    operations = [
        migrations.AlterField(
            model_name='orderanditemcharges',
            name='name',
            field=models.CharField(db_index=True, max_length=128, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='weightband',
            name='upper_limit',
            field=models.DecimalField(db_index=True, decimal_places=3, help_text='Enter upper limit of this weight band in kg. The lower limit will be determined by the other weight bands.', max_digits=12, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Upper Limit'),
        ),
        migrations.AlterField(
            model_name='weightbased',
            name='name',
            field=models.CharField(db_index=True, max_length=128, unique=True, verbose_name='Name'),
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-01 07:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('payment', '0004_auto_20181115_1953'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='source',
            options={'ordering': ['pk'], 'verbose_name': 'Source', 'verbose_name_plural': 'Sources'},
        ),
        migrations.AlterModelOptions(
            name='sourcetype',
            options={'ordering': ['name'], 'verbose_name': 'Source Type', 'verbose_name_plural': 'Source Types'},
        ),
        migrations.AlterField(
            model_name='sourcetype',
            name='name',
            field=models.CharField(db_index=True, max_length=128, verbose_name='Name'),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class OrderConfig(OscarConfig):
    label = "order"
    name = "oscar.apps.order"
    verbose_name = _("Order")



from django.forms.models import BaseModelFormSet, modelformset_factory
from django.utils.functional import cached_property

from oscar.core.loading import get_classes, get_model

Line = get_model("basket", "line")
BasketLineForm, SavedLineForm = get_classes(
    "basket.forms", ["BasketLineForm", "SavedLineForm"]
)


class BaseBasketLineFormSet(BaseModelFormSet):
    def __init__(self, strategy, *args, **kwargs):
        self.strategy = strategy
        super().__init__(*args, **kwargs)

    def _construct_form(self, i, **kwargs):
        return super()._construct_form(i, strategy=self.strategy, **kwargs)

    def _should_delete_form(self, form):
        """
        Quantity of zero is treated as if the user checked the DELETE checkbox,
        which results in the basket line being deleted
        """
        if super()._should_delete_form(form):
            return True

        # If related form instance already removed, let's remove this form
        # as well.
        if not form.instance.id:
            return True
        if self.can_delete and "quantity" in form.cleaned_data:  # pylint: disable=E1101
            return form.cleaned_data["quantity"] == 0

    @cached_property
    def forms_with_instances(self):
        return [f for f in self.forms if f.instance.id]

    def __iter__(self):
        """
        Skip forms with removed lines when iterating through the formset.
        """
        return iter(self.forms_with_instances)


BasketLineFormSet = modelformset_factory(
    Line, form=BasketLineForm, formset=BaseBasketLineFormSet, extra=0, can_delete=True
)


class BaseSavedLineFormSet(BaseModelFormSet):
    def __init__(self, strategy, basket, *args, **kwargs):
        self.strategy = strategy
        self.basket = basket
        super().__init__(*args, **kwargs)

    def _construct_form(self, i, **kwargs):
        return super()._construct_form(
            i, strategy=self.strategy, basket=self.basket, **kwargs
        )


SavedLineFormSet = modelformset_factory(
    Line, form=SavedLineForm, formset=BaseSavedLineFormSet, extra=0, can_delete=True
)



# Generated by Django 2.1.7 on 2019-03-01 10:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0007_auto_20181115_1953'),
        ('communication', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='communicationevent',
            name='event_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='communication.CommunicationEventType', verbose_name='Event Type'),
        ),
    ]



# This is a rewrite of django-sorting but with added support for i18n title
# strings.
# See https://github.com/directeur/django-sorting

from django import template
from django.conf import settings
from django.utils.html import format_html
from django.utils.safestring import mark_safe

register = template.Library()

DEFAULT_SORT_UP = getattr(
    settings, "DEFAULT_SORT_UP", '<i class="fas fa-chevron-up"></i>'
)
DEFAULT_SORT_DOWN = getattr(
    settings, "DEFAULT_SORT_DOWN", '<i class="fas fa-chevron-down"></i>'
)

sort_directions = {
    "asc": {"icon": DEFAULT_SORT_UP, "inverse": "desc"},
    "desc": {"icon": DEFAULT_SORT_DOWN, "inverse": "asc"},
    "": {"icon": DEFAULT_SORT_UP, "inverse": "desc"},
}


@register.simple_tag(takes_context=True)
def anchor(context, field, title=None):
    field = field.strip()
    if title is None:
        title = field.capitalize()
    title = title.strip()
    request = context["request"]
    get_vars = request.GET.copy()
    sort_field = get_vars.pop("sort", [None])[0]

    icon = ""
    if sort_field == field:
        # We are already sorting on this field, so we set the inverse
        # direction within the GET params that get used within the href.
        direction = get_vars.pop("dir", [""])[0]
        get_vars["dir"] = sort_directions[direction]["inverse"]
        icon = sort_directions[direction]["icon"]
    icon = mark_safe(icon)

    href = "%s?sort=%s" % (request.path, field)
    if len(get_vars) > 0:
        href += "&%s" % get_vars.urlencode()
    if icon:
        title = format_html("{} {}", title, icon)
    return format_html('<a href="{}">{}</a>', mark_safe(href), title)



from django.utils.translation import ngettext_lazy
from django_tables2 import Table


class DashboardTable(Table):
    caption = ngettext_lazy("%d Row", "%d Rows")

    def get_caption_display(self):
        # Allow overriding the caption with an arbitrary string that we cannot
        # interpolate the number of rows in
        try:
            return self.caption % self.paginator.count
        except TypeError:
            pass
        return self.caption

    class Meta:
        template_name = "oscar/dashboard/table.html"
        attrs = {"class": "table table-striped table-bordered"}



from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ReportGenerator = get_class("dashboard.reports.reports", "ReportGenerator")
ReportCSVFormatter = get_class("dashboard.reports.reports", "ReportCSVFormatter")
ReportHTMLFormatter = get_class("dashboard.reports.reports", "ReportHTMLFormatter")
Voucher = get_model("voucher", "Voucher")


class VoucherReportCSVFormatter(ReportCSVFormatter):
    filename_template = "voucher-performance.csv"

    def generate_csv(self, response, vouchers):
        writer = self.get_csv_writer(response)
        header_row = [
            _("Voucher code"),
            _("Added to a basket"),
            _("Used in an order"),
            _("Total discount"),
        ]
        writer.writerow(header_row)

        for voucher in vouchers:
            row = [
                voucher.code,
                voucher.num_basket_additions,
                voucher.num_orders,
                voucher.total_discount,
            ]
            writer.writerow(row)


class VoucherReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "oscar/dashboard/reports/partials/voucher_report.html"


class VoucherReportGenerator(ReportGenerator):
    code = "vouchers"
    description = _("Voucher performance")
    model_class = Voucher

    formatters = {
        "CSV_formatter": VoucherReportCSVFormatter,
        "HTML_formatter": VoucherReportHTMLFormatter,
    }



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-03-23 15:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('payment', '0002_auto_20141007_2032'),
    ]

    operations = [
        migrations.AlterField(
            model_name='source',
            name='reference',
            field=models.CharField(blank=True, max_length=255, verbose_name='Reference'),
        ),
    ]



from decimal import Decimal as D

import factory

from oscar.core.loading import get_model

__all__ = [
    "PartnerFactory",
    "StockRecordFactory",
]


class PartnerFactory(factory.django.DjangoModelFactory):
    name = "Gardners"

    class Meta:
        model = get_model("partner", "Partner")

    @factory.post_generation
    def users(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for user in extracted:
                self.users.add(user)  # pylint: disable=E1101


class StockRecordFactory(factory.django.DjangoModelFactory):
    partner = factory.SubFactory(PartnerFactory)
    partner_sku = factory.Sequence(lambda n: "unit%d" % n)
    price_currency = "GBP"
    price = D("9.99")
    num_in_stock = 100

    class Meta:
        model = get_model("partner", "StockRecord")



from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

OSCAR_SHOP_NAME = "Oscar"
OSCAR_SHOP_TAGLINE = ""
OSCAR_HOMEPAGE = reverse_lazy("catalogue:index")

# Dynamic class loading
OSCAR_DYNAMIC_CLASS_LOADER = "oscar.core.loading.default_class_loader"

# Basket settings
OSCAR_BASKET_COOKIE_LIFETIME = 7 * 24 * 60 * 60
OSCAR_BASKET_COOKIE_OPEN = "oscar_open_basket"
OSCAR_BASKET_COOKIE_SECURE = False
OSCAR_MAX_BASKET_QUANTITY_THRESHOLD = 10000

# Recently-viewed products
OSCAR_RECENTLY_VIEWED_COOKIE_LIFETIME = 7 * 24 * 60 * 60
OSCAR_RECENTLY_VIEWED_COOKIE_NAME = "oscar_history"
OSCAR_RECENTLY_VIEWED_COOKIE_SECURE = False
OSCAR_RECENTLY_VIEWED_PRODUCTS = 20

# Currency
OSCAR_DEFAULT_CURRENCY = "GBP"

# Paths
OSCAR_IMAGE_FOLDER = "images/products/%Y/%m/"
OSCAR_DELETE_IMAGE_FILES = True

# Copy this image from oscar/static/img to your MEDIA_ROOT folder.
# It needs to be there so Sorl can resize it.
OSCAR_MISSING_IMAGE_URL = "image_not_found.jpg"

# Address settings
OSCAR_REQUIRED_ADDRESS_FIELDS = (
    "first_name",
    "last_name",
    "line1",
    "line4",
    "postcode",
    "country",
)

# Pagination settings

OSCAR_OFFERS_PER_PAGE = 20
OSCAR_PRODUCTS_PER_PAGE = 20
OSCAR_REVIEWS_PER_PAGE = 20
OSCAR_NOTIFICATIONS_PER_PAGE = 20
OSCAR_EMAILS_PER_PAGE = 20
OSCAR_ORDERS_PER_PAGE = 20
OSCAR_ADDRESSES_PER_PAGE = 20
OSCAR_STOCK_ALERTS_PER_PAGE = 20
OSCAR_DASHBOARD_ITEMS_PER_PAGE = 20

# Checkout
OSCAR_ALLOW_ANON_CHECKOUT = False

# Reviews
OSCAR_ALLOW_ANON_REVIEWS = True
OSCAR_MODERATE_REVIEWS = False

# Accounts
OSCAR_ACCOUNTS_REDIRECT_URL = "customer:profile-view"

# This enables sending alert notifications/emails instantly when products get
# back in stock by listening to stock record update signals.
# This might impact performance for large numbers of stock record updates.
# Alternatively, the management command ``oscar_send_alerts`` can be used to
# run periodically, e.g. as a cron job. In this case eager alerts should be
# disabled.
OSCAR_EAGER_ALERTS = True

# Registration
OSCAR_SEND_REGISTRATION_EMAIL = True
OSCAR_FROM_EMAIL = "oscar@example.com"

# Slug handling
OSCAR_SLUG_FUNCTION = "oscar.core.utils.default_slugifier"
OSCAR_SLUG_MAP = {}
OSCAR_SLUG_BLACKLIST = []
OSCAR_SLUG_ALLOW_UNICODE = False

# Cookies
OSCAR_COOKIES_DELETE_ON_LOGOUT = [
    "oscar_recently_viewed_products",
]

# Offers
OSCAR_OFFERS_INCL_TAX = False
# Values (using the names of the model constants) from
# "offer.ConditionalOffer.TYPE_CHOICES"
OSCAR_OFFERS_IMPLEMENTED_TYPES = [
    "SITE",
    "VOUCHER",
]

# Hidden Oscar features, e.g. wishlists or reviews
OSCAR_HIDDEN_FEATURES = []

# Menu structure of the dashboard navigation
OSCAR_DASHBOARD_NAVIGATION = [
    {
        "label": _("Dashboard"),
        "icon": "fas fa-list",
        "url_name": "dashboard:index",
    },
    {
        "label": _("Catalogue"),
        "icon": "fas fa-sitemap",
        "children": [
            {
                "label": _("Products"),
                "url_name": "dashboard:catalogue-product-list",
            },
            {
                "label": _("Product Types"),
                "url_name": "dashboard:catalogue-class-list",
            },
            {
                "label": _("Categories"),
                "url_name": "dashboard:catalogue-category-list",
            },
            {
                "label": _("Ranges"),
                "url_name": "dashboard:range-list",
            },
            {
                "label": _("Low stock alerts"),
                "url_name": "dashboard:stock-alert-list",
            },
            {
                "label": _("Options"),
                "url_name": "dashboard:catalogue-option-list",
            },
            {
                "label": _("Attribute Option Groups"),
                "url_name": "dashboard:catalogue-attribute-option-group-list",
            },
        ],
    },
    {
        "label": _("Fulfilment"),
        "icon": "fas fa-shopping-cart",
        "children": [
            {
                "label": _("Orders"),
                "url_name": "dashboard:order-list",
            },
            {
                "label": _("Statistics"),
                "url_name": "dashboard:order-stats",
            },
            {
                "label": _("Partners"),
                "url_name": "dashboard:partner-list",
            },
            # The shipping method dashboard is disabled by default as it might
            # be confusing. Weight-based shipping methods aren't hooked into
            # the shipping repository by default (as it would make
            # customising the repository slightly more difficult).
            # {
            #     'label': _('Shipping charges'),
            #     'url_name': 'dashboard:shipping-method-list',
            # },
        ],
    },
    {
        "label": _("Customers"),
        "icon": "fas fa-users",
        "children": [
            {
                "label": _("Customers"),
                "url_name": "dashboard:users-index",
            },
            {
                "label": _("Stock alert requests"),
                "url_name": "dashboard:user-alert-list",
            },
        ],
    },
    {
        "label": _("Offers"),
        "icon": "fas fa-bullhorn",
        "children": [
            {
                "label": _("Offers"),
                "url_name": "dashboard:offer-list",
            },
            {
                "label": _("Vouchers"),
                "url_name": "dashboard:voucher-list",
            },
            {
                "label": _("Voucher Sets"),
                "url_name": "dashboard:voucher-set-list",
            },
        ],
    },
    {
        "label": _("Content"),
        "icon": "fas fa-folder",
        "children": [
            {
                "label": _("Pages"),
                "url_name": "dashboard:page-list",
            },
            {
                "label": _("Email templates"),
                "url_name": "dashboard:comms-list",
            },
            {
                "label": _("Reviews"),
                "url_name": "dashboard:reviews-list",
            },
        ],
    },
    {
        "label": _("Reports"),
        "icon": "fas fa-chart-bar",
        "url_name": "dashboard:reports-index",
    },
]
OSCAR_DASHBOARD_DEFAULT_ACCESS_FUNCTION = "oscar.apps.dashboard.nav.default_access_fn"

# Search facets
OSCAR_SEARCH_FACETS = {
    "fields": {
        # The key for these dicts will be used when passing facet data
        # to the template. Same for the 'queries' dict below.
        "product_class": {"name": _("Type"), "field": "product_class"},
        "rating": {"name": _("Rating"), "field": "rating"},
        # You can specify an 'options' element that will be passed to the
        # SearchQuerySet.facet() call.
        # For instance, with Elasticsearch backend, 'options': {'order': 'term'}
        # will sort items in a facet by title instead of number of items.
        # It's hard to get 'missing' to work
        # correctly though as of Solr's hilarious syntax for selecting
        # items without a specific facet:
        # http://wiki.apache.org/solr/SimpleFacetParameters#facet.method
        # 'options': {'missing': 'true'}
    },
    "queries": {
        "price_range": {
            "name": _("Price range"),
            "field": "price",
            "queries": [
                # This is a list of (name, query) tuples where the name will
                # be displayed on the front-end.
                (_("0 to 20"), "[0 TO 20]"),
                (_("20 to 40"), "[20 TO 40]"),
                (_("40 to 60"), "[40 TO 60]"),
                (_("60+"), "[60 TO *]"),
            ],
        },
    },
}

OSCAR_THUMBNAILER = "oscar.core.thumbnails.SorlThumbnail"

OSCAR_URL_SCHEMA = "http"

OSCAR_SAVE_SENT_EMAILS_TO_DB = True

OSCAR_CATALOGUE_USE_POSTGRES_MATERIALISED_VIEWS = False

HAYSTACK_SIGNAL_PROCESSOR = "haystack.signals.RealtimeSignalProcessor"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0004_auto_20150217_1710'),
    ]

    operations = [
        migrations.AlterField(
            model_name='product',
            name='product_class',
            field=models.ForeignKey(related_name='products', on_delete=django.db.models.deletion.PROTECT, blank=True, to='catalogue.ProductClass', help_text='Choose what type of product this is', null=True, verbose_name='Product type'),
        ),
    ]



import logging

from django.core.management.base import BaseCommand
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class

logger = logging.getLogger(__name__)

AlertsDispatcher = get_class("customer.alerts.utils", "AlertsDispatcher")


class Command(BaseCommand):
    """
    Check stock records of products for availability and send out alerts
    to customers that have registered for an alert.
    """

    help = _("Check for products that are back in stock and send out alerts")

    def handle(self, *args, **options):
        """
        Check all products with active product alerts for
        availability and send out email alerts when a product is
        available to buy.
        """
        AlertsDispatcher().send_alerts()



from decimal import Decimal as D

from oscar.core.loading import get_class
from oscar.test import factories

Default = get_class("partner.strategy", "Default")


def add_product(basket, price=None, quantity=1, product=None):
    """
    Helper to add a product to the basket.
    """
    has_strategy = False
    try:
        has_strategy = hasattr(basket, "strategy")
    except RuntimeError:
        pass
    if not has_strategy:
        basket.strategy = Default()
    if price is None:
        price = D("1")
    if product and product.has_stockrecords:
        record = product.stockrecords.first()
    else:
        record = factories.create_stockrecord(
            product=product, price=price, num_in_stock=quantity + 1
        )
    basket.add_product(record.product, quantity)


def add_products(basket, args):
    """
    Helper to add a series of products to the passed basket
    """
    for price, quantity in args:
        add_product(basket, price, quantity)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
from decimal import Decimal
import oscar.models.fields
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Benefit',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(verbose_name='Type', max_length=128, blank=True, choices=[('Percentage', "Discount is a percentage off of the product's value"), ('Absolute', "Discount is a fixed amount off of the basket's total"), ('Multibuy', 'Discount is to give the cheapest product for free'), ('Fixed price', 'Get the products that meet the condition for a fixed price'), ('Shipping absolute', 'Discount is a fixed amount of the shipping cost'), ('Shipping fixed price', 'Get shipping for a fixed price'), ('Shipping percentage', 'Discount is a percentage off of the shipping cost')])),
                ('value', oscar.models.fields.PositiveDecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Value', null=True)),
                ('max_affected_items', models.PositiveIntegerField(verbose_name='Max Affected Items', blank=True, help_text='Set this to prevent the discount consuming all items within the range that are in the basket.', null=True)),
                ('proxy_class', oscar.models.fields.NullCharField(unique=True, verbose_name='Custom class', default=None, max_length=255)),
            ],
            options={
                'verbose_name_plural': 'Benefits',
                'verbose_name': 'Benefit',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Condition',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(verbose_name='Type', max_length=128, blank=True, choices=[('Count', 'Depends on number of items in basket that are in condition range'), ('Value', 'Depends on value of items in basket that are in condition range'), ('Coverage', 'Needs to contain a set number of DISTINCT items from the condition range')])),
                ('value', oscar.models.fields.PositiveDecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Value', null=True)),
                ('proxy_class', oscar.models.fields.NullCharField(unique=True, verbose_name='Custom class', default=None, max_length=255)),
            ],
            options={
                'verbose_name_plural': 'Conditions',
                'verbose_name': 'Condition',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ConditionalOffer',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(verbose_name='Name', unique=True, max_length=128, help_text="This is displayed within the customer's basket")),
                ('slug', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Slug', max_length=128, editable=False, blank=True)),
                ('description', models.TextField(verbose_name='Description', help_text='This is displayed on the offer browsing page', blank=True)),
                ('offer_type', models.CharField(default='Site', max_length=128, verbose_name='Type', choices=[('Site', 'Site offer - available to all users'), ('Voucher', 'Voucher offer - only available after entering the appropriate voucher code'), ('User', 'User offer - available to certain types of user'), ('Session', 'Session offer - temporary offer, available for a user for the duration of their session')])),
                ('status', models.CharField(default='Open', max_length=64, verbose_name='Status')),
                ('priority', models.IntegerField(default=0, verbose_name='Priority', help_text='The highest priority offers are applied first')),
                ('start_datetime', models.DateTimeField(blank=True, verbose_name='Start date', null=True)),
                ('end_datetime', models.DateTimeField(verbose_name='End date', blank=True, help_text="Offers are active until the end of the 'end date'", null=True)),
                ('max_global_applications', models.PositiveIntegerField(verbose_name='Max global applications', blank=True, help_text='The number of times this offer can be used before it is unavailable', null=True)),
                ('max_user_applications', models.PositiveIntegerField(verbose_name='Max user applications', blank=True, help_text='The number of times a single user can use this offer', null=True)),
                ('max_basket_applications', models.PositiveIntegerField(verbose_name='Max basket applications', blank=True, help_text='The number of times this offer can be applied to a basket (and order)', null=True)),
                ('max_discount', models.DecimalField(verbose_name='Max discount', max_digits=12, decimal_places=2, null=True, help_text='When an offer has given more discount to orders than this threshold, then the offer becomes unavailable', blank=True)),
                ('total_discount', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Total Discount')),
                ('num_applications', models.PositiveIntegerField(default=0, verbose_name='Number of applications')),
                ('num_orders', models.PositiveIntegerField(default=0, verbose_name='Number of Orders')),
                ('redirect_url', oscar.models.fields.ExtendedURLField(verbose_name='URL redirect (optional)', blank=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('benefit', models.ForeignKey(verbose_name='Benefit', to='offer.Benefit', on_delete=models.CASCADE)),
                ('condition', models.ForeignKey(verbose_name='Condition', to='offer.Condition', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-priority'],
                'verbose_name_plural': 'Conditional offers',
                'verbose_name': 'Conditional offer',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Range',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=128, verbose_name='Name')),
                ('slug', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Slug', max_length=128, editable=False, blank=True)),
                ('description', models.TextField(verbose_name="Description", blank=True)),
                ('is_public', models.BooleanField(default=False, verbose_name='Is public?', help_text='Public ranges have a customer-facing page')),
                ('includes_all_products', models.BooleanField(default=False, verbose_name='Includes all products?')),
                ('proxy_class', oscar.models.fields.NullCharField(unique=True, verbose_name='Custom class', default=None, max_length=255)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('classes', models.ManyToManyField(related_name='classes', verbose_name='Product Types', to='catalogue.ProductClass', blank=True)),
                ('excluded_products', models.ManyToManyField(related_name='excludes', verbose_name='Excluded Products', to='catalogue.Product', blank=True)),
                ('included_categories', models.ManyToManyField(related_name='includes', verbose_name='Included Categories', to='catalogue.Category', blank=True)),
            ],
            options={
                'verbose_name_plural': 'Ranges',
                'verbose_name': 'Range',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RangeProduct',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('display_order', models.IntegerField(default=0)),
                ('product', models.ForeignKey(to='catalogue.Product', on_delete=models.CASCADE)),
                ('range', models.ForeignKey(to='offer.Range', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RangeProductFileUpload',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('filepath', models.CharField(max_length=255, verbose_name='File Path')),
                ('size', models.PositiveIntegerField(verbose_name='Size')),
                ('date_uploaded', models.DateTimeField(auto_now_add=True, verbose_name='Date Uploaded')),
                ('status', models.CharField(default='Pending', max_length=32, verbose_name='Status', choices=[('Pending', 'Pending'), ('Failed', 'Failed'), ('Processed', 'Processed')])),
                ('error_message', models.CharField(max_length=255, verbose_name='Error Message', blank=True)),
                ('date_processed', models.DateTimeField(verbose_name='Date Processed', null=True)),
                ('num_new_skus', models.PositiveIntegerField(verbose_name='Number of New SKUs', null=True)),
                ('num_unknown_skus', models.PositiveIntegerField(verbose_name='Number of Unknown SKUs', null=True)),
                ('num_duplicate_skus', models.PositiveIntegerField(verbose_name='Number of Duplicate SKUs', null=True)),
                ('range', models.ForeignKey(verbose_name='Range', related_name='file_uploads', to='offer.Range', on_delete=models.CASCADE)),
                ('uploaded_by', models.ForeignKey(verbose_name='Uploaded By', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('-date_uploaded',),
                'verbose_name_plural': 'Range Product Uploaded Files',
                'verbose_name': 'Range Product Uploaded File',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='rangeproduct',
            unique_together=set([('range', 'product')]),
        ),
        migrations.AddField(
            model_name='range',
            name='included_products',
            field=models.ManyToManyField(related_name='includes', verbose_name='Included Products', to='catalogue.Product', through='offer.RangeProduct', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='condition',
            name='range',
            field=models.ForeignKey(null=True, verbose_name='Range', to='offer.Range', blank=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='benefit',
            name='range',
            field=models.ForeignKey(null=True, verbose_name='Range', to='offer.Range', blank=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.CreateModel(
            name='AbsoluteDiscountBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Absolute discount benefits',
                'verbose_name': 'Absolute discount benefit',
                'proxy': True,
            },
            bases=('offer.benefit',),
        ),
        migrations.CreateModel(
            name='CountCondition',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Count conditions',
                'verbose_name': 'Count condition',
                'proxy': True,
            },
            bases=('offer.condition',),
        ),
        migrations.CreateModel(
            name='CoverageCondition',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Coverage Conditions',
                'verbose_name': 'Coverage Condition',
                'proxy': True,
            },
            bases=('offer.condition',),
        ),
        migrations.CreateModel(
            name='FixedPriceBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Fixed price benefits',
                'verbose_name': 'Fixed price benefit',
                'proxy': True,
            },
            bases=('offer.benefit',),
        ),
        migrations.CreateModel(
            name='MultibuyDiscountBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Multibuy discount benefits',
                'verbose_name': 'Multibuy discount benefit',
                'proxy': True,
            },
            bases=('offer.benefit',),
        ),
        migrations.CreateModel(
            name='PercentageDiscountBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Percentage discount benefits',
                'verbose_name': 'Percentage discount benefit',
                'proxy': True,
            },
            bases=('offer.benefit',),
        ),
        migrations.CreateModel(
            name='ShippingBenefit',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('offer.benefit',),
        ),
        migrations.CreateModel(
            name='ShippingAbsoluteDiscountBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Shipping absolute discount benefits',
                'verbose_name': 'Shipping absolute discount benefit',
                'proxy': True,
            },
            bases=('offer.shippingbenefit',),
        ),
        migrations.CreateModel(
            name='ShippingFixedPriceBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Fixed price shipping benefits',
                'verbose_name': 'Fixed price shipping benefit',
                'proxy': True,
            },
            bases=('offer.shippingbenefit',),
        ),
        migrations.CreateModel(
            name='ShippingPercentageDiscountBenefit',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Shipping percentage discount benefits',
                'verbose_name': 'Shipping percentage discount benefit',
                'proxy': True,
            },
            bases=('offer.shippingbenefit',),
        ),
        migrations.CreateModel(
            name='ValueCondition',
            fields=[
            ],
            options={
                'verbose_name_plural': 'Value conditions',
                'verbose_name': 'Value condition',
                'proxy': True,
            },
            bases=('offer.condition',),
        ),
    ]



# Generated by Django 3.0.8 on 2020-08-01 07:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0019_option_required'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='option',
            options={'ordering': ['name'], 'verbose_name': 'Option', 'verbose_name_plural': 'Options'},
        ),
        migrations.AlterField(
            model_name='option',
            name='name',
            field=models.CharField(db_index=True, max_length=128, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='option',
            name='required',
            field=models.BooleanField(default=False, verbose_name='Is this option required?'),
        ),
    ]



import zlib
from decimal import Decimal as D
from operator import itemgetter

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Sum
from django.utils.encoding import smart_str
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from oscar.core.compat import AUTH_USER_MODEL
from oscar.core.loading import get_class, get_classes
from oscar.core.utils import (
    get_default_currency,
    round_half_up_two_dec,
    round_half_up_four_dec,
)
from oscar.core.decorators import deprecated
from oscar.models.fields.slugfield import SlugField
from oscar.templatetags.currency_filters import currency

OfferApplications = get_class("offer.results", "OfferApplications")
Unavailable = get_class("partner.availability", "Unavailable")
LineDiscountRegistry = get_class("basket.utils", "LineDiscountRegistry")
OpenBasketManager, SavedBasketManager = get_classes(
    "basket.managers", ["OpenBasketManager", "SavedBasketManager"]
)


class AbstractBasket(models.Model):
    """
    Basket object
    """

    # Baskets can be anonymously owned - hence this field is nullable.  When a
    # anon user signs in, their two baskets are merged.
    owner = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True,
        related_name="baskets",
        on_delete=models.CASCADE,
        verbose_name=_("Owner"),
    )

    # Basket statuses
    # - Frozen is for when a basket is in the process of being submitted
    #   and we need to prevent any changes to it.
    OPEN, MERGED, SAVED, FROZEN, SUBMITTED = (
        "Open",
        "Merged",
        "Saved",
        "Frozen",
        "Submitted",
    )
    STATUS_CHOICES = (
        (OPEN, _("Open - currently active")),
        (MERGED, _("Merged - superceded by another basket")),
        (SAVED, _("Saved - for items to be purchased later")),
        (FROZEN, _("Frozen - the basket cannot be modified")),
        (SUBMITTED, _("Submitted - has been ordered at the checkout")),
    )
    status = models.CharField(
        _("Status"), max_length=128, default=OPEN, choices=STATUS_CHOICES
    )

    # A basket can have many vouchers attached to it.  However, it is common
    # for sites to only allow one voucher per basket - this will need to be
    # enforced in the project's codebase.
    vouchers = models.ManyToManyField(
        "voucher.Voucher", verbose_name=_("Vouchers"), blank=True
    )

    date_created = models.DateTimeField(_("Date created"), auto_now_add=True)
    date_merged = models.DateTimeField(_("Date merged"), null=True, blank=True)
    date_submitted = models.DateTimeField(_("Date submitted"), null=True, blank=True)

    # Only if a basket is in one of these statuses can it be edited
    editable_statuses = (OPEN, SAVED)

    class Meta:
        abstract = True
        app_label = "basket"
        verbose_name = _("Basket")
        verbose_name_plural = _("Baskets")

    objects = models.Manager()
    open = OpenBasketManager()
    saved = SavedBasketManager()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We keep a cached copy of the basket lines as we refer to them often
        # within the same request cycle.  Also, applying offers will append
        # discount data to the basket lines which isn't persisted to the DB and
        # so we want to avoid reloading them as this would drop the discount
        # information.
        self._lines = None
        self.offer_applications = OfferApplications()

    def __str__(self):
        return _("%(status)s basket (owner: %(owner)s, lines: %(num_lines)d)") % {
            "status": self.status,
            "owner": self.owner,
            "num_lines": self.num_lines,
        }

    # ========
    # Strategy
    # ========

    @property
    def has_strategy(self):
        return hasattr(self, "_strategy")

    def _get_strategy(self):
        if not self.has_strategy:
            raise RuntimeError(
                "No strategy class has been assigned to this basket. "
                "This is normally assigned to the incoming request in "
                "oscar.apps.basket.middleware.BasketMiddleware. "
                "Since it is missing, you must be doing something different. "
                "Ensure that a strategy instance is assigned to the basket!"
            )
        return self._strategy

    def _set_strategy(self, strategy):
        self._strategy = strategy  # pylint: disable=W0201

    strategy = property(_get_strategy, _set_strategy)

    def all_lines(self):
        """
        Return a cached set of basket lines.

        This is important for offers as they alter the line models and you
        don't want to reload them from the DB as that information would be
        lost.
        """
        if self.id is None:
            return self.lines.model.objects.none()  # pylint: disable=E1101
        if self._lines is None:
            self._lines = (
                self.lines.select_related("product", "stockrecord")
                .prefetch_related("attributes", "product__images")
                .order_by(self._meta.pk.name)
            )
        return self._lines

    def max_allowed_quantity(self):
        """
        Returns maximum product quantity, that can be added to the basket
        with the respect to basket quantity threshold.
        """
        basket_threshold = settings.OSCAR_MAX_BASKET_QUANTITY_THRESHOLD
        if basket_threshold:
            total_basket_quantity = self.num_items
            max_allowed = basket_threshold - total_basket_quantity
            return max_allowed, basket_threshold
        return None, None

    def is_quantity_allowed(self, qty, line=None):
        """
        Test whether the passed quantity of items can be added to the basket
        """
        # We enforce a max threshold to prevent a DOS attack via the offers
        # system.
        max_allowed, basket_threshold = self.max_allowed_quantity()

        if line is not None:
            (
                line_purchase_permitted,
                reason,
            ) = line.purchase_info.availability.is_purchase_permitted(qty)

            if not line_purchase_permitted:
                return line_purchase_permitted, reason

            # Also check if it's permitted with potentional other lines of the same product & stocrecord
            total_lines_quantity = self.basket_quantity(line) + qty
            (
                line_purchase_permitted,
                reason,
            ) = line.purchase_info.availability.is_purchase_permitted(
                total_lines_quantity
            )

            if not line_purchase_permitted:
                return line_purchase_permitted, _(
                    "Available stock is only %(max)d, which has been exceeded because "
                    "multiple lines contain the same product."
                ) % {"max": line.purchase_info.availability.num_available}

        if max_allowed is not None and qty > max_allowed:
            return False, _(
                "Due to technical limitations we are not able "
                "to ship more than %(threshold)d items in one order."
            ) % {"threshold": basket_threshold}

        return True, None

    def basket_quantity(self, line):
        """Return the quantity of similar lines in the basket.
        The basket can contain multiple lines with the same product and
        stockrecord, but different options. Those quantities are summed up.
        """
        if self.id is None:
            return 0
        matching_lines = self.lines.filter(stockrecord=line.stockrecord)
        quantity = matching_lines.aggregate(Sum("quantity"))["quantity__sum"]
        return quantity or 0

    # ============
    # Manipulation
    # ============

    def flush(self):
        """
        Remove all lines from basket.
        """
        if self.status == self.FROZEN:
            raise PermissionDenied("A frozen basket cannot be flushed")
        self.lines.all().delete()
        self._lines = None

    # pylint: disable=unused-argument
    def get_stock_info(self, product, options):
        """
        Hook for implementing strategies that depend on product options
        """
        # The built-in strategies don't use options, so initially disregard
        # them.
        return self.strategy.fetch_for_product(product)

    def add_product(self, product, quantity=1, options=None):
        """
        Add a product to the basket

        The 'options' list should contains dicts with keys 'option' and 'value'
        which link the relevant product.Option model and string value
        respectively.

        Returns (line, created).
          line: the matching basket line
          created: whether the line was created or updated

        """
        if options is None:
            options = []
        if not self.id:
            self.save()

        # Ensure that all lines are the same currency
        price_currency = self.currency
        stock_info = self.get_stock_info(product, options)

        if not stock_info.price.exists:
            raise ValueError("Strategy hasn't found a price for product %s" % product)

        if price_currency and stock_info.price.currency != price_currency:
            raise ValueError(
                (
                    "Basket lines must all have the same currency. Proposed "
                    "line has currency %s, while basket has currency %s"
                )
                % (stock_info.price.currency, price_currency)
            )

        if stock_info.stockrecord is None:
            raise ValueError(
                (
                    "Basket lines must all have stock records. Strategy hasn't "
                    "found any stock record for product %s"
                )
                % product
            )

        # Line reference is used to distinguish between variations of the same
        # product (eg T-shirts with different personalisations)
        line_ref = self._create_line_reference(product, stock_info.stockrecord, options)

        # Determine price to store (if one exists).  It is only stored for
        # audit and sometimes caching.
        defaults = {
            "quantity": max(0, quantity),
            "price_excl_tax": stock_info.price.excl_tax,
            "price_currency": stock_info.price.currency,
            "tax_code": stock_info.price.tax_code,
        }
        if stock_info.price.is_tax_known:
            defaults["price_incl_tax"] = stock_info.price.incl_tax

        line, created = self.lines.get_or_create(
            line_reference=line_ref,
            product=product,
            stockrecord=stock_info.stockrecord,
            defaults=defaults,
        )
        if created:
            for option_dict in options:
                line.attributes.create(
                    option=option_dict["option"], value=option_dict["value"]
                )
        else:
            line.quantity = max(0, line.quantity + quantity)
            line.save()
        self.reset_offer_applications()

        # Returning the line is useful when overriding this method.
        return line, created

    add_product.alters_data = True
    add = add_product

    def applied_offers(self):
        """
        Return a dict of offers successfully applied to the basket.

        This is used to compare offers before and after a basket change to see
        if there is a difference.
        """
        return self.offer_applications.offers

    def reset_offer_applications(self):
        """
        Remove any discounts so they get recalculated
        """
        self.offer_applications = OfferApplications()
        self._lines = None

    def merge_line(self, line, add_quantities=True):
        """
        For transferring a line from another basket to this one.

        This is used with the "Saved" basket functionality.
        """
        try:
            existing_line = self.lines.get(line_reference=line.line_reference)
        except ObjectDoesNotExist:
            # Line does not already exist - reassign its basket
            line.basket = self
            line.save()
        else:
            # Line already exists - assume the max quantity is correct and
            # delete the old
            if add_quantities:
                existing_line.quantity += line.quantity
            else:
                existing_line.quantity = max(existing_line.quantity, line.quantity)
            existing_line.save()
            line.delete()
        finally:
            self._lines = None

    merge_line.alters_data = True

    def merge(self, basket, add_quantities=True):
        """
        Merges another basket with this one.

        :basket: The basket to merge into this one.
        :add_quantities: Whether to add line quantities when they are merged.
        """
        # Use basket.lines.all instead of all_lines as this function is called
        # before a strategy has been assigned.
        for line_to_merge in basket.lines.all():
            self.merge_line(line_to_merge, add_quantities)
        basket.status = self.MERGED
        basket.date_merged = now()
        basket._lines = None
        basket.save()
        # Ensure all vouchers are moved to the new basket
        for voucher in basket.vouchers.all():
            basket.vouchers.remove(voucher)
            self.vouchers.add(voucher)

    merge.alters_data = True

    def freeze(self):
        """
        Freezes the basket so it cannot be modified.
        """
        self.status = self.FROZEN
        self.save()

    freeze.alters_data = True

    def thaw(self):
        """
        Unfreezes a basket so it can be modified again
        """
        self.status = self.OPEN
        self.save()

    thaw.alters_data = True

    def submit(self):
        """
        Mark this basket as submitted
        """
        self.status = self.SUBMITTED
        self.date_submitted = now()
        self.save()

    submit.alters_data = True

    # Kept for backwards compatibility
    set_as_submitted = submit

    def is_shipping_required(self):
        """
        Test whether the basket contains physical products that require
        shipping.
        """
        for line in self.all_lines():
            if line.product.is_shipping_required:
                return True
        return False

    # =======
    # Helpers
    # =======

    def _create_line_reference(self, product, stockrecord, options):
        """
        Returns a reference string for a line based on the item
        and its options.
        """
        base = "%s_%s" % (product.id, stockrecord.id)
        if not options:
            return base
        repr_options = [
            {"option": repr(option["option"]), "value": repr(option["value"])}
            for option in options
        ]
        repr_options.sort(key=itemgetter("option"))
        return "%s_%s" % (base, zlib.crc32(repr(repr_options).encode("utf8")))

    def _get_total(self, model_property):
        """
        For executing a named method on each line of the basket
        and returning the total.
        """
        total = D("0.00")
        for line in self.all_lines():
            try:
                total += getattr(line, model_property)
            except ObjectDoesNotExist:
                # Handle situation where the product may have been deleted
                pass
            except TypeError:
                # Handle Unavailable products with no known price
                info = self.get_stock_info(line.product, line.attributes.all())
                if info.availability.is_available_to_buy:
                    raise
        return round_half_up_two_dec(total)

    # ==========
    # Properties
    # ==========

    @property
    def is_empty(self):
        """
        Test if this basket is empty
        """
        return self.id is None or self.num_lines == 0

    @property
    def is_tax_known(self):
        """
        Test if tax values are known for this basket.

        If the basket is empty, then tax values are unknown.
        """
        return (not self.is_empty) and all(
            [line.is_tax_known for line in self.all_lines()]
        )

    @property
    def total_excl_tax(self):
        """
        Return total line price excluding tax
        """
        return self._get_total("line_price_excl_tax_incl_discounts")

    @property
    def total_tax(self):
        """Return total tax for a line"""
        return self._get_total("line_tax")

    @property
    def total_incl_tax(self):
        """
        Return total price inclusive of tax and discounts
        """
        return self._get_total("line_price_incl_tax_incl_discounts")

    @property
    def total_incl_tax_excl_discounts(self):
        """
        Return total price inclusive of tax but exclusive discounts
        """
        return self._get_total("line_price_incl_tax")

    @property
    def total_discount(self):
        return self._get_total("discount_value")

    @property
    def offer_discounts(self):
        """
        Return basket discounts from non-voucher sources.  Does not include
        shipping discounts.
        """
        return self.offer_applications.offer_discounts

    @property
    def voucher_discounts(self):
        """
        Return discounts from vouchers
        """
        return self.offer_applications.voucher_discounts

    @property
    def has_shipping_discounts(self):
        return len(self.shipping_discounts) > 0

    @property
    def shipping_discounts(self):
        """
        Return discounts from vouchers
        """
        return self.offer_applications.shipping_discounts

    @property
    def post_order_actions(self):
        """
        Return discounts from vouchers
        """
        return self.offer_applications.post_order_actions

    @property
    def grouped_voucher_discounts(self):
        """
        Return discounts from vouchers but grouped so that a voucher which
        links to multiple offers is aggregated into one object.
        """
        return self.offer_applications.grouped_voucher_discounts

    @property
    def total_excl_tax_excl_discounts(self):
        """
        Return total price excluding tax and discounts
        """
        return self._get_total("line_price_excl_tax")

    @property
    def num_lines(self):
        """Return number of lines"""
        return self.all_lines().count()

    @property
    def num_items(self):
        """Return number of items"""
        return sum(line.quantity for line in self.all_lines())

    @property
    def num_items_without_discount(self):
        num = 0
        for line in self.all_lines():
            num += line.quantity_without_discount
        return num

    @property
    def num_items_with_discount(self):
        num = 0
        for line in self.all_lines():
            num += line.quantity_with_discount
        return num

    @property
    def time_before_submit(self):
        if not self.date_submitted:
            return None
        return self.date_submitted - self.date_created

    @property
    def time_since_creation(self, test_datetime=None):
        if not test_datetime:
            test_datetime = now()
        return test_datetime - self.date_created

    @property
    def contains_a_voucher(self):
        if self.id is None:
            return False
        return self.vouchers.exists()

    @property
    def is_submitted(self):
        return self.status == self.SUBMITTED

    @property
    def can_be_edited(self):
        """
        Test if a basket can be edited
        """
        return self.status in self.editable_statuses

    @property
    def currency(self):
        # Since all lines should have the same currency, return the currency of
        # the first one found.
        for line in self.all_lines():
            return line.price_currency

    # =============
    # Query methods
    # =============

    def contains_voucher(self, code):
        """
        Test whether the basket contains a voucher with a given code
        """
        if self.id is None:
            return False
        try:
            self.vouchers.get(code=code)
        except ObjectDoesNotExist:
            return False
        else:
            return True

    def product_quantity(self, product):
        """
        Return the quantity of a product in the basket

        The basket can contain multiple lines with the same product, but
        different options and stockrecords. Those quantities are summed up.
        """
        if self.id is None:
            return 0
        matching_lines = self.lines.filter(product=product)
        quantity = matching_lines.aggregate(Sum("quantity"))["quantity__sum"]
        return quantity or 0

    def line_quantity(self, product, stockrecord, options=None):
        """
        Return the current quantity of a specific product and options
        """
        if self.id is None:
            return 0
        ref = self._create_line_reference(product, stockrecord, options)
        try:
            return self.lines.get(line_reference=ref).quantity
        except ObjectDoesNotExist:
            return 0


class AbstractLine(models.Model):
    """A line of a basket (product and a quantity)

    Common approaches on ordering basket lines:

        a) First added at top. That's the history-like approach; new items are
           added to the bottom of the list. Changing quantities doesn't impact
           position.
           Oscar does this by default. It just sorts by Line.pk, which is
           guaranteed to increment after each creation.

        b) Last modified at top. That means items move to the top when you add
           another one, and new items are added to the top as well.  Amazon
           mostly does this, but doesn't change the position when you update
           the quantity in the basket view.
           To get this behaviour, change Meta.ordering and optionally do
           something similar on wishlist lines. Order lines should already
           be created in the order of the basket lines, and are sorted by
           their primary key, so no changes should be necessary there.

    """

    basket = models.ForeignKey(
        "basket.Basket",
        on_delete=models.CASCADE,
        related_name="lines",
        verbose_name=_("Basket"),
    )

    # This is to determine which products belong to the same line
    # We can't just use product.id as you can have customised products
    # which should be treated as separate lines.  Set as a
    # SlugField as it is included in the path for certain views.
    line_reference = SlugField(_("Line Reference"), max_length=128, db_index=True)

    product = models.ForeignKey(
        "catalogue.Product",
        on_delete=models.CASCADE,
        related_name="basket_lines",
        verbose_name=_("Product"),
    )

    # We store the stockrecord that should be used to fulfil this line.
    stockrecord = models.ForeignKey(
        "partner.StockRecord", on_delete=models.CASCADE, related_name="basket_lines"
    )

    quantity = models.PositiveIntegerField(_("Quantity"), default=1)

    # We store the unit price incl tax of the product when it is first added to
    # the basket.  This allows us to tell if a product has changed price since
    # a person first added it to their basket.
    price_currency = models.CharField(
        _("Currency"), max_length=12, default=get_default_currency
    )
    price_excl_tax = models.DecimalField(
        _("Price excl. Tax"), decimal_places=2, max_digits=12, null=True
    )
    price_incl_tax = models.DecimalField(
        _("Price incl. Tax"), decimal_places=2, max_digits=12, null=True
    )
    tax_code = models.CharField(
        _("VAT rate code"), max_length=64, blank=True, null=True
    )

    # Track date of first addition
    date_created = models.DateTimeField(
        _("Date Created"), auto_now_add=True, db_index=True
    )
    date_updated = models.DateTimeField(_("Date Updated"), auto_now=True, db_index=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.discounts = LineDiscountRegistry(self)

    class Meta:
        abstract = True
        app_label = "basket"
        # Enforce sorting by order of creation.
        ordering = ["date_created", "pk"]
        unique_together = ("basket", "line_reference")
        verbose_name = _("Basket line")
        verbose_name_plural = _("Basket lines")

    @property
    @deprecated
    def consumer(self):
        return self.discounts

    def __str__(self):
        return _(
            "Basket #%(basket_id)d, Product #%(product_id)d, quantity %(quantity)d"
        ) % {
            "basket_id": self.basket.pk,
            "product_id": self.product.pk,
            "quantity": self.quantity,
        }

    def save(self, *args, **kwargs):
        if not self.basket.can_be_edited:
            raise PermissionDenied(
                _("You cannot modify a %s basket") % (self.basket.status.lower(),)
            )
        return super().save(*args, **kwargs)

    # =============
    # Offer methods
    # =============

    def clear_discount(self):
        """
        Remove any discounts from this line.
        """
        self.discounts = LineDiscountRegistry(self)

    def discount(self, discount_value, affected_quantity, incl_tax=True, offer=None):
        """
        Apply a discount to this line
        """
        if incl_tax:
            if self.discounts.excl_tax > 0:
                raise RuntimeError(
                    "Attempting to discount the tax-inclusive price of a line "
                    "when tax-exclusive discounts are already applied"
                )
        else:
            if self.discounts.incl_tax > 0:
                raise RuntimeError(
                    "Attempting to discount the tax-exclusive price of a line "
                    "when tax-inclusive discounts are already applied"
                )
        self.discounts.discount(discount_value, affected_quantity, incl_tax, offer)

    def consume(self, quantity, offer=None):
        """
        Mark all or part of the line as 'consumed'

        Consumed items are no longer available to be used in offers.
        """
        return self.discounts.consume(quantity, offer=offer)

    def get_price_breakdown(self):
        """
        Return a breakdown of line prices after discounts have been applied.

        Returns a list of (unit_price_incl_tax, unit_price_excl_tax, quantity)
        tuples.
        """
        if not self.is_tax_known:
            raise RuntimeError(
                "A price breakdown can only be determined when taxes are known"
            )
        prices = []
        if not self.discount_value:
            prices.append(
                (self.unit_price_incl_tax, self.unit_price_excl_tax, self.quantity)
            )
        else:
            # Need to split the discount among the affected quantity
            # of products.
            item_incl_tax_discount = self.discount_value / int(
                self.discounts.num_consumed()
            )
            item_excl_tax_discount = item_incl_tax_discount * self._tax_ratio
            item_excl_tax_discount = round_half_up_four_dec(item_excl_tax_discount)
            prices.append(
                (
                    self.unit_price_incl_tax - item_incl_tax_discount,
                    self.unit_price_excl_tax - item_excl_tax_discount,
                    self.discounts.num_consumed(),
                )
            )
            if self.quantity_without_discount:
                prices.append(
                    (
                        self.unit_price_incl_tax,
                        self.unit_price_excl_tax,
                        self.quantity_without_discount,
                    )
                )
        return prices

    # =======
    # Helpers
    # =======

    @property
    def _tax_ratio(self):
        # this function tries to computate the tax ratio based on the incl tax price
        # versus the excl tax price.
        if not self.unit_price_incl_tax:
            return 0
        return self.unit_price_excl_tax / self.unit_price_incl_tax

    # ===============
    # Offer Discounts
    # ===============

    def has_offer_discount(self, offer):
        return self.discounts.num_consumed(offer) > 0

    def quantity_with_offer_discount(self, offer):
        return self.discounts.num_consumed(offer)

    def quantity_without_offer_discount(self, offer):
        return self.discounts.available(offer)

    def is_available_for_offer_discount(self, offer):
        return self.discounts.available(offer) > 0

    def quantity_available_for_offer(self, offer):
        return self.quantity_without_offer_discount(
            offer
        ) + self.quantity_with_offer_discount(offer)

    # ==========
    # Properties
    # ==========

    @property
    @deprecated
    def _discount_incl_tax(self):
        return self.discounts.incl_tax

    @_discount_incl_tax.setter
    def _discount_incl_tax(self, value):
        raise Exception("You can not set _discount_incl_tax")

    @property
    @deprecated
    def _discount_excl_tax(self):
        return self.discounts.excl_tax

    @_discount_excl_tax.setter
    def _discount_excl_tax(self, value):
        raise Exception("You can not set _discount_excl_tax")

    @property
    def has_discount(self):
        return bool(self.discounts.num_consumed())

    @property
    def quantity_with_discount(self):
        return self.discounts.num_consumed()

    @property
    def quantity_without_discount(self):
        return self.discounts.available()

    @property
    def discount_value(self):
        return self.discounts.total

    # pylint: disable=W0201
    @property
    def purchase_info(self):
        """
        Return the stock/price info
        """
        if not hasattr(self, "_info"):
            # Cache the PurchaseInfo instance.
            self._info = self.basket.strategy.fetch_for_line(self, self.stockrecord)
        return self._info

    @property
    def is_tax_known(self):
        return self.purchase_info.price.is_tax_known

    @property
    def unit_effective_price(self):
        """
        The price to use for offer calculations
        """
        return self.purchase_info.price.effective_price

    @property
    def unit_price_excl_tax(self):
        return self.purchase_info.price.excl_tax

    @property
    def unit_price_incl_tax(self):
        return self.purchase_info.price.incl_tax

    @property
    def unit_tax(self):
        return self.purchase_info.price.tax

    @property
    def line_price_excl_tax(self):
        if self.unit_price_excl_tax is not None:
            return self.quantity * self.unit_price_excl_tax

    @property
    def line_price_excl_tax_incl_discounts(self):
        if self.line_price_excl_tax is None:
            return None

        excl_tax_discounts = self.discounts.excl_tax
        if excl_tax_discounts:
            # these are discounts that return a value excluding tax, we can simply
            # subtract this value from line_price_excl_tax to get to line_price_excl_tax_incl_discounts
            return max(0, self.line_price_excl_tax - excl_tax_discounts)

        # This is a tricky situation.  We know the discount as calculated
        # against tax inclusive prices but we need to guess how much of the
        # discount applies to tax-exclusive prices.  We do this by
        # assuming a linear tax and scaling down the original discount.

        incl_tax_discounts = self.discounts.incl_tax
        if incl_tax_discounts and self._tax_ratio:
            if self.line_price_excl_tax is not None:
                return max(
                    0,
                    round_half_up_two_dec(
                        self.line_price_excl_tax
                        - (self._tax_ratio * incl_tax_discounts)
                    ),
                )
            elif self.line_price_incl_tax is not None:
                # when all else fails, compute based on line_price_incl_tax
                return max(
                    0, self._tax_ratio * (self.line_price_incl_tax - incl_tax_discounts)
                )

        # there are no discounts so just return the line_price_excl_tax
        return self.line_price_excl_tax

    @property
    def line_price_incl_tax_incl_discounts(self):
        excl_tax_discounts = self.discounts.excl_tax
        incl_tax_discounts = self.discounts.incl_tax

        # We use whichever discount value is set.  If the discount value was
        # calculated against the tax-exclusive prices, then the line price
        # including tax
        if self.line_price_incl_tax is not None and incl_tax_discounts:
            return max(0, self.line_price_incl_tax - incl_tax_discounts)
        elif self.line_price_excl_tax is not None and excl_tax_discounts:
            return round_half_up_two_dec(
                self.line_price_excl_tax_incl_discounts / self._tax_ratio
            )

        return self.line_price_incl_tax

    @property
    def line_tax(self):
        if self.is_tax_known:
            return (
                self.line_price_incl_tax_incl_discounts
                - self.line_price_excl_tax_incl_discounts
            )

    @property
    def line_price_incl_tax(self):
        if self.unit_price_incl_tax is not None:
            return self.quantity * self.unit_price_incl_tax

    @property
    def description(self):
        d = smart_str(self.product)
        ops = []
        for attribute in self.attributes.all():
            value = attribute.value
            if isinstance(value, list):
                ops.append(
                    "%s = '%s'"
                    % (attribute.option.name, (", ".join([str(v) for v in value])))
                )
            else:
                ops.append("%s = '%s'" % (attribute.option.name, value))
        if ops:
            d = "%s (%s)" % (d, ", ".join(ops))
        return d

    def get_warning(self):
        """
        Return a warning message about this basket line if one is applicable

        This could be things like the price has changed
        """
        if isinstance(self.purchase_info.availability, Unavailable):
            return _("'%(product)s' is no longer available") % {
                "product": self.product.get_title()
            }

        if not self.price_incl_tax:
            return
        if not self.purchase_info.price.is_tax_known:
            return

        # Compare current price to price when added to basket
        current_price_incl_tax = self.purchase_info.price.incl_tax
        if current_price_incl_tax != self.price_incl_tax:
            product_prices = {
                "product": self.product.get_title(),
                "old_price": currency(self.price_incl_tax, self.price_currency),
                "new_price": currency(current_price_incl_tax, self.price_currency),
            }
            if current_price_incl_tax > self.price_incl_tax:
                warning = _(
                    "The price of '%(product)s' has increased from"
                    " %(old_price)s to %(new_price)s since you added"
                    " it to your basket"
                )
                return warning % product_prices
            else:
                warning = _(
                    "The price of '%(product)s' has decreased from"
                    " %(old_price)s to %(new_price)s since you added"
                    " it to your basket"
                )
                return warning % product_prices


class AbstractLineAttribute(models.Model):
    """
    An attribute of a basket line
    """

    line = models.ForeignKey(
        "basket.Line",
        on_delete=models.CASCADE,
        related_name="attributes",
        verbose_name=_("Line"),
    )
    option = models.ForeignKey(
        "catalogue.Option", on_delete=models.CASCADE, verbose_name=_("Option")
    )
    value = models.JSONField(_("Value"), encoder=DjangoJSONEncoder)

    class Meta:
        abstract = True
        app_label = "basket"
        verbose_name = _("Line attribute")
        verbose_name_plural = _("Line attributes")



from django.contrib import admin

from oscar.core.loading import get_model

ProductReview = get_model("reviews", "ProductReview")
Vote = get_model("reviews", "Vote")


class ProductReviewAdmin(admin.ModelAdmin):
    list_display = (
        "product",
        "title",
        "score",
        "status",
        "total_votes",
        "delta_votes",
        "date_created",
    )
    readonly_fields = ("total_votes", "delta_votes")


class VoteAdmin(admin.ModelAdmin):
    list_display = ("review", "user", "delta", "date_created")


admin.site.register(ProductReview, ProductReviewAdmin)
admin.site.register(Vote, VoteAdmin)



from django import template
from django.contrib.sessions.serializers import JSONSerializer

from oscar.core.loading import get_class, get_model

AddToBasketForm = get_class("basket.forms", "AddToBasketForm")
SimpleAddToBasketForm = get_class("basket.forms", "SimpleAddToBasketForm")
Product = get_model("catalogue", "product")

register = template.Library()

QNT_SINGLE, QNT_MULTIPLE = "single", "multiple"


@register.simple_tag
def basket_form(request, product, quantity_type="single"):
    if not isinstance(product, Product):
        return ""

    initial = {}
    if not product.is_parent:
        initial["product_id"] = product.id

    form_class = AddToBasketForm
    if quantity_type == QNT_SINGLE:
        form_class = SimpleAddToBasketForm

    basket_post_data = request.session.pop(
        "add_to_basket_form_post_data_%s" % product.pk, None
    )

    if basket_post_data is not None:
        basket_post_data = JSONSerializer().loads(basket_post_data.encode("latin-1"))

    form = form_class(
        request.basket, data=basket_post_data, product=product, initial=initial
    )

    return form



# pylint: disable=W0201

from django.apps import apps
from django.conf import settings
from django.urls import path, reverse_lazy
from django.views.generic.base import RedirectView

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class Shop(OscarConfig):
    name = "oscar"

    def ready(self):
        from django.contrib.auth.forms import SetPasswordForm
        import oscar.checks  # pylint: disable=W0611

        self.catalogue_app = apps.get_app_config("catalogue")
        self.customer_app = apps.get_app_config("customer")
        self.basket_app = apps.get_app_config("basket")
        self.checkout_app = apps.get_app_config("checkout")
        self.search_app = apps.get_app_config("search")
        self.dashboard_app = apps.get_app_config("dashboard")
        self.offer_app = apps.get_app_config("offer")
        self.wishlists_app = apps.get_app_config("wishlists")

        self.password_reset_form = get_class("customer.forms", "PasswordResetForm")
        self.set_password_form = SetPasswordForm

    def get_urls(self):
        from django.contrib.auth import views as auth_views

        from oscar.views.decorators import login_forbidden

        urls = [
            path("", RedirectView.as_view(url=settings.OSCAR_HOMEPAGE), name="home"),
            path("catalogue/", self.catalogue_app.urls),
            path("basket/", self.basket_app.urls),
            path("checkout/", self.checkout_app.urls),
            path("accounts/", self.customer_app.urls),
            path("search/", self.search_app.urls),
            path("dashboard/", self.dashboard_app.urls),
            path("offers/", self.offer_app.urls),
            path("wishlists/", self.wishlists_app.urls),
            # Password reset - as we're using Django's default view functions,
            # we can't namespace these urls as that prevents
            # the reverse function from working.
            path(
                "password-reset/",
                login_forbidden(
                    auth_views.PasswordResetView.as_view(
                        form_class=self.password_reset_form,
                        success_url=reverse_lazy("password-reset-done"),
                        template_name="oscar/registration/password_reset_form.html",
                    )
                ),
                name="password-reset",
            ),
            path(
                "password-reset/done/",
                login_forbidden(
                    auth_views.PasswordResetDoneView.as_view(
                        template_name="oscar/registration/password_reset_done.html"
                    )
                ),
                name="password-reset-done",
            ),
            path(
                "password-reset/confirm/<str:uidb64>/<str:token>/",
                login_forbidden(
                    auth_views.PasswordResetConfirmView.as_view(
                        form_class=self.set_password_form,
                        success_url=reverse_lazy("password-reset-complete"),
                        template_name="oscar/registration/password_reset_confirm.html",
                    )
                ),
                name="password-reset-confirm",
            ),
            path(
                "password-reset/complete/",
                login_forbidden(
                    auth_views.PasswordResetCompleteView.as_view(
                        template_name="oscar/registration/password_reset_complete.html"
                    )
                ),
                name="password-reset-complete",
            ),
        ]
        return urls
